import sys,glob,re,os
from tokenize import String
import numpy as np
import fnmatch
import xes_neo.gui.xes_data
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.figure import Figure
import copy
from xes_neo.gui.xes_individual import *
from xes_neo.gui.xes_fit import *
import scipy as scipy

"""
Author: Evan Restuccia evan@restuccias.com
"""


def sort_fold_list(dirs):
    fold_list = list_dirs(dirs)
    fold_list.sort(key=natural_keys)
    return fold_list
## Human Sort
def list_dirs(path):
    return [os.path.basename(x) for x in filter(
        os.path.isdir, glob.glob(os.path.join(path, '*')))]


def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    '''
    return [ atoi(c) for c in re.split(r'(\d+)', text) ]




class xes_analysis():
    """
    Contains all functions for the reading of and analysis of output files from the actualy genetic algorithm
    """
    def __init__(self,dirs,params, peakType, data_KE = False, data_XES = False, scale_var = False):
        """
        Params(Dictionary) : an dictionary of parameters

        Note, self.bkgns and self.peaks and lists of options for the possible types of peaks vs. backgrounds. This will allow the
        file reading portion to tell what is a background and what is a peak

        self.data_obj is an instance of the XES_Data class
        self.is_singlet will become an array of bools where true = singlet, false = doublet

        """
        self.dirs = dirs
        fileName = params['fileName']
        self.bkgns = ['Shirley-Sherwood','Baseline','SVSC_shirley','Exponential', 'Linear', 'Polynomial 1', 'Polynomial 2', 'Polynomial 3', '3-Param Tougaard', '2-Param Tougaard']
        self.peaks = params['peaks']
        self.peak_options = self.peaks
        self.peakType = peakType
       
        for i in range(len(self.peaks)):
            self.peak_options[i] = self.peak_options[i].lower()
        self.bkgn_options = self.bkgns
        #self.numPeaks = len(self.peaks)
        for i in range(len(self.bkgns)):
            self.bkgns[i] = self.bkgns[i].lower()

        for i in range(len(self.peaks)):
            self.peaks[i] = self.peaks[i].lower()


        self.data_obj = params['data obj']
        self.is_singlet = []
        self.is_coster_kronig = []


        self.x = self.data_obj.get_x(data_KE, data_XES)
        self.y = self.data_obj.get_y(scale_var)
       

        #number of parameters that the GA fit
        #self.number_of_parameters = self.calculate_number_of_params(self.peaks,self.bkgns)

    def get_param_num(self,type):
        """
        (helper) takes in a peak/background type and matches it with the number of parameters it has
        always has the number of parameters + 1, the +1 is for it's type

        """
        
        #Peak Types:
        if type == 'voigt':
            return 5
        if type == 'gaussian':
            return 4
        if type == 'lorentzian':
            return 4
        if type == 'double lorentzian':
            return 6
        if type == 'doniach-sunjic':
            return 6

        #Backgrounds:
        if type == 'shirley-sherwood':

            return 2
        if type == 'exponential':
            return 3
        if type == 'linear':
            return 3
        if type == 'svsc_shirley': #Not actually being used right now --> takes way too long to run
            return 2
        if type == 'baseline':
            return 2
        if type == 'polynomial 1':
            return 1
        if type == 'polynomial 2':
            return 1
        if type == 'polynomial 3':
            return 1
        if type == '3-param tougaard':
            return 3
        if type == '2-param tougaard':
            return 1
        else:
            print("Warning, need to add parameter number to get_param_num in analysis file")
            print("Type found was:" + type)
        
    def calculate_number_of_params(self,types):
        """
        peaks(array) array containing strings which represent the type of each peak, length is equal to the number of peaks
        bkgns(array) length is number of backgrounds, each element is a string representing the type of background

        (helper)
        Adds the number of parameters each type has together so the analysis will know how many columns to have in arrays of the parameters
        """

        num = 0
        for i,one_type in enumerate(types):
            #voigt has 4 params
            num += self.get_param_num(one_type.lower().strip())


        return num

    def extract_data(self,scale_var,plot_err=True,passin=''):
        
        scale_var = scale_var
        
        """
        Extract data value using array data

        plot_err(bool) whether or not to plot the error
        passin(num?) Optional- the best fit

        After extracting data, evaluates all individuals and assigns self.best_ind to the best chisqr individual
        """
        #gets the parameters and best fit if they are not passed in
        if passin == '':
            full_param_list = []
            files = []
            for r, d, f in os.walk(self.dirs):
                f.sort(key = natural_keys)
                for file in f:
                    if fnmatch.fnmatch(file,'*_data.csv'):
                        files.append(os.path.join(r, file))
            files.sort(key=natural_keys)

            for i in range(len(files)):
                final_values,self.types = self.extract_from_result_file(files[i])
                full_param_list.append(final_values)
                if i == 0:
                    stack = full_param_list[i]
                stack = np.vstack((stack,full_param_list[i]))

            for i in range(len(self.types)): #Try selecting an output folder
                self.types[i]= self.types[i].strip()

            self.number_of_parameters = self.calculate_number_of_params(self.types)
            #print(full_param_list)
            self.full_matrix = stack
            #print(stack)
            bestFit,err = self.construct_bestfit_err_mat(stack,scale_var,plot_err)

            best_Fit = np.reshape(copy.deepcopy(bestFit),(1, -1))

            #print(best_Fit)
        else:
            err = np.zeros([self.number_of_parameters,1])
            err = passin
            bestFit = np.zeros([self.number_of_parameters,1])
            best_Fit = passin

        self.bestFit = best_Fit.flatten()
        self.err = err.flatten()
        #should print error here later, but need to customize it to number of params
        self.err = list(self.err)
        new_err = []
        new_best = []
        for i in range(len(self.err)):
            new_err.append(round(self.err[i],2))
            new_best.append(round(self.bestFit[i],2))
        #print("Parameter Error values are: ", new_err)



        error_array = []
        peaks = []
        bkgns = []
        bkgn_len_mod = 0
        bkgn_len_mod = 0
        for i,Type in enumerate(self.types):
            if Type.lower().strip() in self.peaks:
                peaks.append(Type)
            if Type.lower().strip() in self.bkgns:
                bkgns.append(Type)

                if Type == 'Linear':
                    bkgn_len_mod +=0

                if Type == '3-Param Tougaard':
                    bkgn_len_mod +=1
        error_length = self.number_of_parameters - len(peaks) - len(bkgns)

        #print(new_err)

        #Since error values come in one large array, I created a series of if and for statements that split up the array into the seperate peaks and sums them into an array of arrays
        #The background errors go to their own array
        l = 0
       
        k = 0
        for peak in peaks: #iterates for as many peaks there are
            #Fixed error in logic --> Should work for all peaks now 11/06/2023
            #Issues when using two different peak types: Need to fix this, if two different peaks change range of second peak to (l, k+1)
            error = []
            #print("PEAK", peak)
            #print("PEAK", peak)

            #Issue with error values and coster-kronig: All go to zero? Need to fix
            if(peak.lower() == 'voigt'):
               
                k += 4




                if self.is_singlet[0] == False:
                    k += 2
                
                for j in np.arange(l, k):
                    error.append(new_err[j])
              
                l = k

            elif(peak.lower() == 'gaussian'):
               
                
                k += 3
               
                #print(k, l)


                if self.is_singlet[0] == False:
                    k += 2 #added -2 in for issue --> Not sure why it gets too big. Logic error
                    #print("doublet", k, l)
               
                for j in np.arange(l, k):
                    #print(k, j)
                    #print("ERROR", new_err[j])
                    #print(k, j)
                    #print("ERROR", new_err[j])
                    error.append(new_err[j])
               
                l = k

            elif(peak.lower() == 'lorentzian'):
                k += 3




                if self.is_singlet[0] == False:
                    k += 2
               
                for j in np.arange(l, k):
                    error.append(new_err[j])

                l = k

            elif(peak.lower() == 'double lorentzian'):
                
                k += 5
               
                


                if self.is_singlet[0] == False:
                    k += 2
              
                for j in np.arange(l, k):
                    #print(k, j)

                    #print("ERROR", new_err[j])
                    error.append(new_err[j])
                    if new_err[j] == 0.0:
                        pass

                l = k

            elif(peak.lower() == 'doniach-sunjic'):
                k += 5


                if self.is_singlet[0] == False:
                    k += 2
              
                for j in np.arange(l, k):
                    error.append(new_err[j])

                l = k

           
            error_array.append(error)
            

        error_bkgns = []
        bkgn_len = len(new_err) - len(bkgns) - bkgn_len_mod
        bkgn_len = len(new_err) - len(bkgns) - bkgn_len_mod
        print(new_err)
        for i in reversed(range(bkgn_len, len(new_err))):
                error_bkgns.append(new_err[i])
        error_bkgns = error_bkgns[::-1] #flip it to appear in the same order as the printed out peak attributes
        self.error_bkgns = error_bkgns
        self.error_array = error_array





        self.bestFit_mat = best_Fit

        # print("Not Normalized:")
        self.best_Fit_n = copy.deepcopy(best_Fit)
        #print(self.best_Fit_n)
        # self.best_Fit_n[:,0] -= self.scaler.min_
        # self.best_Fit_n[:,0] /= self.scaler.scale_
        # print(self.best_Fit_n)
        # self.x_model = np.arange
        
        self.best_ind = self.create_individual_from_bestfit(self.best_Fit_n[0],scale_var)
        print("peak attributes read are:")
        #print(self.best_Fit_n)
        print(self.best_ind.get_params(self.is_coster_kronig))
       
        self.y_model,self.peak_components,self.bkgn_components = self.best_ind.getFitWithComponents(self.x,self.y, self.peakType, self.bkgns)

        #Moved print statement down so that errors appear after the attributes:
        print("Peak Error Values are: ", error_array)
        print("Background Error Values are: ", error_bkgns)
        '''
        self.y_model_components = self.best_ind.get_peaks()
        self.bkgn_components = self.best_ind.get_backgrounds()
        self.totalbackground = [0] * len(self.x)
        for i in range(len(self.bkgn_components)):
            self.bkgn_components[i] = self.bkgn_components[i].getY(self.x,self.y)
            for k in range(len(self.x)):
                self.totalbackground[k] += self.bkgn_components[i][k]

        for i in range(len(self.y_model_components)):
            self.y_model_components[i] = self.y_model_components[i].getY(self.x) + self.totalbackground
        '''

    def extract_from_result_file(self,file):
        """
        Pulls the data from each given file in the selected folder

        When reading these files, it first generates an array of all items in the CSV
        Then it will check the type of the object read.

        When finished, this function will have updated self.types, self.is_singlet, and return csv_numbers
        self.types = array where each element is the type of the next item in the file, i.e.
            1,2,3,4,true,0,'voigt',.5,'SVSC_Shirley' would read as 'voigt','SVSC_Shirley
        self.is_singlet is the same as self.types except it only has the booleans
        csv_numbers is all the numbers left over

        The reasoning for 3 different arrays is the original analysis uses reshape often, which
        requires all the same type

        """
        try:
            os.path.exists(file)
            csv_unflatten = []
            csv_numbers = []
            gen_csv = np.genfromtxt(file,delimiter=',',dtype = None,encoding= None)
            for row in gen_csv:
                
                temp_row = []
                temp_numbers = []
                types = []
                is_doublet = False
                peaks = []
                
                for k in range(len(row)):
                    typeCheck = type(row[k])
                    if(typeCheck == np.float64):
                        temp_numbers.append(np.float64(row[k]))
                    elif(typeCheck == np.str_):
                        #types.append(row[k])
                        types.append(row[k])
                    elif(typeCheck == np.bool_):
                        
                        #Added this condition in since we now have Coster-Kronig which is bool
                       
                        
                        self.is_singlet.append(row[k])
                           
                        self.is_coster_kronig.append(row[k])
                       
                        #ADDED IN THE FIXED BUTTON. I THINK THIS IS MESSING IT UP --> Adds in 4 extra BOOLS

                        
                    temp_row.append(row[k])

                #ADD STATEMENT CONIDTION FOR IF NO BOOL TYPES FOUND

                #Add in condition for if is_singlet == False?



                if len(csv_unflatten) == 0:
                        csv_unflatten = temp_row
                        csv_numbers = temp_numbers
                else:
                    csv_unflatten = np.vstack((csv_unflatten,temp_row))
                    csv_numbers = np.vstack((csv_numbers,temp_numbers))


            coster_kronig = []
            singlet = []
            
            if self.is_singlet[0] == False: #Added this statement here because for some reason when singlet = True it doesnt append coster_kronig = False
                
                for i in range(len(self.is_coster_kronig)): #Coster-Kronig values are every other one in the list appended: Not able to differentiate between Singlet bool and Coster-Kronig
                    if i % 2 == 0: #Even values are Singlet
                        singlet.append(self.is_singlet[i])
                        pass
                    else: #Odd values are Coster-Kronig
                        coster_kronig.append(self.is_coster_kronig[i])
            else:
                for i in range(len(self.is_singlet)):
                    coster_kronig.append(False)
                    singlet = self.is_singlet
            
               
            self.is_coster_kronig = coster_kronig  
            self.is_singlet = singlet
            

                

            if not self.is_coster_kronig:
             
                self.is_coster_kronig.append(False)
            

            #csv_unflatten = gen_csv.reshape((len(gen_csv),self.number_of_parameters+1))
            #print("Type is " + str(gen_csv[0]))
            return csv_numbers,types
        except OSError:
            print(" " + str(file) + " Missing")
            pass

    def  construct_bestfit_err_mat(self,full_matrix,scale_var,plot=False):
        """
        Constructs a matrix of all fits, where each row is an individual and the column is the parameters
        Gets the score of all of these indiiduals, and assigns the bestFit object from that, note that the bestFit is
        just the array of best parameters, not a full individual yet

        full_matrix <matrix>: n by m where n is the number of samples and m is the number of parameters
        npaths <int>: number of paths
        plot <bol>: if it plot or not

        """
        score = []
        full_mat_var_cov = np.cov(full_matrix.T)
        full_mat_diag = np.diag(full_mat_var_cov)
        err = np.sqrt(full_mat_diag)

        labels = self.generate_labels()
#         print(full_mat)
#         bestFit = np.mean(full_mat,axis=0)
#         bestFit = full_mat[-1,:]

        for i in self.full_matrix:
            #arr = i.reshape((1,-1))
            ind = self.create_individual_from_bestfit(i,scale_var)
            score.append(self.fitness(ind))
        bestScore = np.nanargmin(score)
        bestFit = full_matrix[bestScore]
        if plot:
            plt.figure(figsize=(6.4,4.8))
            plt.xticks(np.arange(len(full_mat_diag)),labels[0],rotation=70);
            plt.bar(np.arange(len(full_mat_diag)),np.sqrt(full_mat_diag))

        return bestFit,err

    def create_individual_from_bestfit(self, fit, scale_var):
        """
        fit(list) a list of parameters in order
        First the function creates an empty individual object
        The array uses the self.types list and the get_param_num to get the correct elements of the list,
        and then uses the set function to hand all of those values to each individual peak and/or background as well as with
        its doublet information if is_singlet indicates it is a doublet

        Returns the individual made up of the parameters from the fit
        """
        pars_range = '' #dummy range, since we're going to set the params again later

        #construct peaks and backgrounds list for individual constructor
        peaks = []
        bkgns = []
        j = 0
        count = 0 #Added in to help with multiple shirley backgrounds
        for i,Type in enumerate(self.types):
            if Type.lower().strip() in self.peaks:
                peaks.append(Type)
            if Type.lower().strip() in self.bkgns:
                bkgns.append(Type)
        
        #creates dummy individual, with random attributes
        ind = Individual(bkgns,peaks,scale_var,pars_range)
        peaks2 = ind.get_peaks()
       
        bkgns2 = ind.get_backgrounds()
      
        
        #going to custom set the attributes of the individual to match the ones in the fit
        current_index = 0
        peak_index = 0
        bkgn_index = 0
        
        


        for i,Type in enumerate(self.types):
            num_params = self.get_param_num(Type.lower())
            
            
    
            param_list = []
            doublet_params = []
            CK_params = []
            for k in range(current_index,current_index+num_params-1):
                param_list.append(fit[k])
                
                current_index = k+1

           
            if Type.lower() in self.peak_options:
                if not self.is_singlet[peak_index]:
                    for k in range(current_index, current_index+2):
                        doublet_params.append(fit[k])
                        current_index = k+1
                   
                  
                    if self.is_coster_kronig[peak_index]: #NOT SEEING THIS RIGHT NOW NOT SURE WHY???
  
                        for j in range(current_index, current_index+1):
                            CK_params.append(fit[j])
                            
                            current_index = j + 1
                    
                            
                            



                peaks2[peak_index].set(param_list,doublet_params, CK_params,self.is_singlet[i],self.is_coster_kronig[i])
                peak_index += 1
            

            #At later point in time, we could move this to a single set function in xes fit like the peaks
            if Type.lower() == 'baseline':
                bkgns2[bkgn_index].set_baseline(param_list)
                bkgn_index+=1
                count += 1
            elif Type.lower() == 'shirley-sherwood':
                value =  bkgns2[count].set_shirley_sherwood(param_list) #IDK why but this needs to be here otherwise shirley = -1 when coster-kronig selected 
                peak_num = len(peaks) - 1
                if j <= peak_num:
                    bkgns2[count].set_shirley_sherwood(param_list) #For some reason this line will throw an error when selecting both shirley and linear background type. Shirley has to be selected first for this error to go away
                    j += 1
                    bkgn_index +=1 #Changed from 2 to 1 because I got rid of shirley background parameter since it wasnt doing anything.
                    count += 1
                else:
                    pass
                '''
                if j < len(peaks):
                    j += 1
                    pass
                else:
                    bkgn_index +=2 #changed from 1 to 2 for new Shirley Background using areas
                '''



            elif Type.lower() == 'linear':

                bkgns2[bkgn_index].set_linear(param_list)
                bkgn_index+=2
                count += 1

            elif Type.lower() == '3-param tougaard':

                bkgns2[bkgn_index].set_Tougaard3(param_list)
                bkgn_index+=2
                count += 1



            elif Type.lower() == 'exponential':
                bkgns2[bkgn_index].set_exponential(param_list)
                bkgn_index+=1
                count += 1

            elif Type.lower() == 'svsc_shirley':
                peaks2[peak_index-1].set_svsc_shirley(param_list)
                count += 1

        return ind



    def generate_labels(self):
        """
        I don't know what this one does yet, I never used it
        """
        label=[]
        amp_label = []
        center_label = []
        sigma_label = []
        gamma_label = []
        asymmetry_label = []
        asymmetryDoniach_label = []

        for i in range(1,self.number_of_parameters+1):
            label.append('amp_' + str(i))
            amp_label.append('amp_' + str(i))

            label.append('center_' + str(i))
            center_label.append('center_' + str(i))

            label.append('sigma_' + str(i))
            sigma_label.append('sigma_' + str(i))

            label.append('gamma_' + str(i))
            gamma_label.append('gamma_' + str(i))

            label.append('asymmetry_' + str(i))
            asymmetry_label.append('asymmetry_' + str(i))

            label.append('asymmetryDoniach_' + str(i))
            asymmetryDoniach_label.append('asymmetryDoniach_' + str(i))

        return label,amp_label,center_label,sigma_label,gamma_label,asymmetry_label,asymmetryDoniach_label



    def FWHM(self, indObj):
        FWHM = [0]*len(self.peaks)
        Individual = indObj
        FWHM = Individual.getFWHM(self.x, self.peakType)

        return FWHM



    def fitness(self,indObj):
        """
        Evaluate fitness of a individual
        """
        loss = 0
        Individual = indObj
        N = 30
        y_left = self.y[:N]
        y_right = self.y[-N:]
        self.y_left_avg = sum(y_left)/N
        self.y_right_avg = sum(y_right)/N
        #print(indObj.get_peaks())
        yTotal = np.zeros(len(self.x))
        FWHM = [0]*len(self.peaks)
        self.residual = [0]*len(self.x)
        
        yTotal = Individual.getFit(self.x,self.y, self.peakType, self.bkgns)
        overfit_penalty = 0

        #Segmenting data to see where y falls off 
        
        big_diff = 0

        for i in range(len(self.x)-5):
            y_b4 = self.y[i]
           
            y_after = self.y[i+5]
            
            y_diff = y_b4 - y_after

            if self.x[0] < self.x[-1]: #KE condition

                if y_diff > 0:
                    new_diff = y_diff
                    if new_diff > big_diff:
                        big_diff = new_diff
                        location = i
            else:

                if y_diff < 0:
                    new_diff = y_diff
                    if new_diff < big_diff:
                        big_diff = new_diff
                        location = i

                #print("X B4", self.x[location], "X AFTER", self.x[location+40])
        #print("X high", self.x[location], "X low", self.x[len(self.x)-1])
        y_low_PE_penalty = 0
        for k in range(location, len(self.x)-1):
            if (yTotal[k]- self.y[k]) > 0:
                y_low_PE_penalty += (yTotal[k]- self.y[k])*5000


        for j in range(len(self.x)):
            
            penalty = 0
            bkgn_penalty = 0

            difference = (yTotal[j]- self.y[j])

            if difference > 0:
                overfit_penalty = difference*10000
            sigma = np.sqrt(abs(self.y[j]))

            if sum(yTotal[:N])/N > self.y_left_avg: #Penalty for background going above the data on left side of plot
                diff = sum(yTotal[:N])/N - self.y_left_avg
                bkgn_penalty += 300000*sigma*diff

            if sum(yTotal[:N])/N < self.y_left_avg: #Penalty for background going below the data on left side of plot
                diff = abs(sum(yTotal[:N])/N - self.y_left_avg)
                bkgn_penalty += 300000*sigma*diff

            if sum(yTotal[-N:])/N > self.y_right_avg: #Penalty for background going above the data on right side of plot
                diff = sum(yTotal[-N:])/N - self.y_right_avg
                bkgn_penalty += 30000*sigma*diff

            if sum(yTotal[-N:])/N < self.y_right_avg: #Penalty for going below the data on right side of plot
                diff = abs(sum(yTotal[-N:])/N - self.y_right_avg)
                bkgn_penalty += 300000*sigma*diff

            if difference > 0:
                if difference >= 0.5*sigma:
                    penalty = 1000*sigma

            loss = loss + (difference**2)*sigma + penalty*sigma + bkgn_penalty #+ overfit_penalty + y_low_PE_penalty #This value is different then the value printed out per generation
            self.residual[j] = ((yTotal[j]- self.y[j]))#/(self.y[j])**2
           #loss = loss + ((yTotal[j]- self.y[j])**2)/(self.y[j])**2 #divide by the intensity^2 in order to normalize it amongst other data sets of varying intensities
        #
        #Calculating the error values for the data:
        #Need covariance matrix of all the fitting parameters --> this is dependent upon the peakType
        #covar = np.cov(fit_params) --> takes the covaraince of all the array of fitted parameters
        #residual_error = (yTotal[j]- self.y[j])**2/()
        #res_cov = covar*residual_error
        #for i > len(fit_params):
        #
        #
        # if loss == np.nan:
            # print(individual[0].verbose())
        return loss

    def score(self,verbose=False):
        
        loss =self.fitness(self.best_ind)
        print('Fitness Score (Chi2):',loss)
        # print('Fitness Score (ChiR2):', loss/(len(self.x_raw)-4*self.npaths))
        # self.fwhm(verbose=verbose)
        # self.cal_area(verbose=verbose)
        # print(self.err_full)

    def analyze(self):
        
        #should calculate area and other params
        #print("Full analysis not yet functional")

        def area(yVals):
            """
            (helper) calculate area as Trapezoidal Riemann sum
            """
            area = 0
            #y_total = len(yVals)*[0]
            for i in range(len(yVals)-1):
                """
                (yVals[i]  -  Background[i])  +  (yVals[i + 1]  +  Background[i + 1])
                ----------------------------------------------------------------------
                                (2 * (x[i + 1] - x[i])

                """
                area += (((yVals[i]-self.background[i])+(yVals[i+1]-self.background[i+1]))/2)*(self.x[i+1]-self.x[i])


        
            right_area = 0
            left_area = 0
           
            h = (self.x[-1] -self.x[0])/(len(yVals)-1)
            for i in range(1, len(yVals)):
                right_area += (yVals[i]-self.background[i])*h #Finding area with right endpoint riemann sum (positive/negative area error)
            for i in range(1, len(yVals)):
                left_area += (yVals[i-1]-self.background[i-1])*h #Finding area with left endpoint riemann sum (negative/positive area error)
          
            if left_area < area:
                lower_error = left_area - area
                upper_error = right_area - area
            else:
                lower_error = right_area - area
                upper_error = left_area - area
            
                #y_total[i] = yVals[i]-self.background[i]
            #new_area = scipy.integrate.simps(y_total, self.x) #Calculating area using simpson's rule
            #print("OLD", area, "NEW", new_area, "DIFF", area-new_area) # minimal difference between Simpson's rule and trapozoid rule
            return abs(area), lower_error, upper_error

        self.background = [0] * len(self.x)
        for i in range(len(self.bkgn_components)):
                for j in range(len(self.x)):
                    self.background[j] += self.bkgn_components[i][j]

        self.peak_component_areas = [0] * len(self.peak_components)
        self.peak_upper_errors = [0] * len(self.peak_components)
        self.peak_lower_errors = [0] * len(self.peak_components)
        k = 0
        for peak in self.peak_components:
            self.peak_component_areas[k], self.peak_lower_errors[k], self.peak_upper_errors[k] = area(peak)
            k += 1
            

        self.totalArea, self.lower_error, self.upper_error = area(self.y_model)
        print("Areas")
        print("Total Area Upper and Lower Error:", self.upper_error, self.lower_error)
        print("Total Area: ", self.totalArea)
        print("Individual Peak Area: ", self.peak_component_areas)
        print("Individual Peak Area Upper Errors: ", self.peak_upper_errors)
        print("Individual Peak Area Lower Errors: ", self.peak_lower_errors)
        return self.totalArea, self.peak_component_areas
        #What are the error values for the area? Are they correlated to the background error values???
        '''
        def FWHM_values(yVals):
            MAD = np.mean(np.absolute(yVals-np.mean(self.y))) #Mean Absolute Deviaition
            self.fwhm_g = 2*self.gaussian*np.sqrt(2*np.log(2))
            self.fwhm_l = 2*MAD
            self.fhwm_v = fwhm_l/2 + np.sqrt(pow(fwhm_l, 2)/4 + pow(fwhm_g, 2))
            return self.fwhm_v, self.fwhm_l, self.fwhm_g
        self.FWHM = FWHM_values(self.y_model)
        print(FWHM)
        #FWHM_values = peak.getFWHM(self)
        '''







    def plot_data(self,scale_var,data_KE, data_XES,title='Test',fig_gui=None):
        y_val = []
        if fig_gui == None:
            
            
            first = 1

            y_max = self.y.max()

            #Scaling is done proportional to the maximum peak height. This also sets background to 0
            if y_max <= 0.001:
                scale_val = first*100000000
            elif y_max <= 0.01:
                scale_val = first*10000000
            elif y_max <= 0.1:
                scale_val = first*1000000
            elif y_max <= 1:
                scale_val = first*100000
            elif y_max <= 10:
                scale_val = first*10000
            elif y_max <= 100:
                scale_val = first*1000
            elif y_max <= 1000:
                scale_val = first*100
            elif y_max <= 10000:
                scale_val = first*10
            else:
                scale_val = first

            if scale_var == True:
            
                #y_val = self.y/first #Dividing every element by the first value 
            
                y_val = self.y*scale_val #Multiply by 1000 to scale it
               
            else:
                y_val = self.y
        
            self.y = y_val
                

            
            

            plt.rc('xtick', labelsize='12')
            plt.rc('ytick', labelsize='12')
            # plt.rc('font',size=30)
            plt.rc('figure',autolayout=True)
            plt.rc('axes',titlesize=12,labelsize=12)
            # plt.rc('figure',figsize=(7.2, 4.45))
            plt.rc('axes',linewidth=1)

            plt.rcParams["font.family"] = "Times New Roman"
            # fig,ax = plt.subplots(1,1,figsize=(6,4.5))
            # ax.plot(self.x_raw,self.y_raw,'ko-',linewidth=1,label='Data')
            plt.plot(self.x,self.y,'b--',linewidth=1,label='Data')
            # ax.plot(self.x_slice,self.y_slice,'r--',linewidth=1.2,label='Slice Data')
            plt.plot(self.x,self.y_model,'k',linewidth=1,label='Fit')
            #plt.plot(self.x_linear,self.y_linear,'--',color='tab:purple')
            plt.legend()
        else:
            #ax  = fig_gui.add_subplot(111)
            y_val = []
            first = 1

            y_max = self.y.max()

            #Scaling is done proportional to the maximum peak height. This also sets background to 0
            if y_max <= 0.001:
                scale_val = first*100000000
            elif y_max <= 0.01:
                scale_val = first*10000000
            elif y_max <= 0.1:
                scale_val = first*1000000
            elif y_max <= 1:
                scale_val = first*100000
            elif y_max <= 10:
                scale_val = first*10000
            elif y_max <= 100:
                scale_val = first*1000
            elif y_max <= 1000:
                scale_val = first*100
            elif y_max <= 10000:
                scale_val = first*10
            else:
                scale_val = first
                
            if scale_var == True:
            
                #y_val = self.y/first #Dividing every element by the first value 
            
                y_val = self.y*scale_val #Multiply by 1000 to scale it
              
            else:
                y_val = self.y
            self.y = y_val
          
           #added in gridspec to help make residual plot
            gs = gridspec.GridSpec(2,1, height_ratios=[1,0.1])
            ax = fig_gui.add_subplot(gs[0])
            ax2 = fig_gui.add_subplot(gs[1])
            gs.update(hspace=0)
            ax.plot(self.x,self.y,color='gray', linestyle='None', marker='o', markeredgewidth='0.3', markerfacecolor='None', markeredgecolor='gray', markersize='5',label='Data')
            ax.plot(self.x,self.y_model,'k-',linewidth=1.2,label='Fit')
            ax2.plot(self.x,self.residual,'k-',linewidth=1.2)
            #plot peak componenets
            if(len(self.peak_components)>1):
                for i,peak in enumerate(self.peak_components):

                    ax.plot(self.x,peak,'-',linewidth = 1,label=('Peak ' + str(i+1)))
                    '''
                    if data_KE == False:
                        ax.plot(self.x,peak,'-',linewidth = 1,label=('Peak ' + str(i+1)))
                    else:
                        ax.plot(self.x[::-1],peak,'-',linewidth = 1,label=('Peak ' + str(i+1)))
                    '''

            self.background = [0] * len(self.x)
            for i in range(len(self.bkgn_components)):
                for j in range(len(self.x)):
                    self.background[j] += self.bkgn_components[i][j]
            ax.plot(self.x,self.background,color='dimgray',linestyle='dashed',linewidth =1, label='background')
           
            if data_KE == False:
                ax.invert_xaxis()
                ax2.invert_xaxis()
            else:
                pass
            if data_XES == True:
                ax.invert_xaxis()
                ax2.invert_xaxis()


            #ax.invert_xaxis()
            #ax2.invert_xaxis()
            #ax.plot(self.x_linear,self.y_linear,'--',color='tab:purple')
            ax.legend()
            #ax.set_xlabel("Binding Energy (eV)")
            ax.set_ylabel("Intensity (a.u.)")
            ax2.set_ylabel("Res.")
            if data_KE == False:
                ax2.set_xlabel("Binding Energy (eV)")
            else:
                ax2.set_xlabel("Kinetic Energy (eV)")

            if data_XES == True:
                ax2.set_xlabel("Photon Energy (eV)")
            #ax2.set_xlabel("Binding Energy (eV)")
            fig_gui.tight_layout()

    def get_params(self):
        #return self.params
        is_coster_kronig = self.is_coster_kronig
        parameters = self.best_ind.get_params(is_coster_kronig)
        errors = self.error_array
        print("errors", errors)
        errors_bkgns = self.error_bkgns
        upper_errors_peak_area = self.peak_upper_errors
        lower_errors_peak_area = self.peak_lower_errors
        peak_areas = self.peak_component_areas
        FWHM_values = self.FWHM(self.best_ind)
        peak_y_vals = self.peak_components
        totalFit = self.y_model
        background_fit = self.background
        residual = self.residual
        y_raw = self.y
        diff = 0
        for i in range(len(self.x)):
            diff += pow((totalFit[i] - y_raw[i]),2) 
        res = 1 #1 eV resolution for easyXES100
        step_size = self.x[2] - self.x[1]
        res_step = res/step_size
        N = len(self.x) 
        N_ind = N/res_step
        N_p = len(errors[0])*len(peak_areas) + len(errors_bkgns)
        chi2_r = (N_ind/((N_ind - N_p)*N))*diff
        print("Reduced Chi2: ", round(chi2_r,2))
        return parameters,errors, errors_bkgns, peak_areas, FWHM_values, peak_y_vals, totalFit, background_fit, residual, y_raw, upper_errors_peak_area, lower_errors_peak_area



# Parser for Inputs files
"""
Renamed to parser2 because it was overshadowing the stdlib parser when importing it non-relative for dev stuff- Evan 7/7/23
"""
import configparser
from xes_neo.helper import bcolors

def CheckKey(dict, key_list):
    for i in range(len(key_list)):
        try:
            dict[key_list[i]]

        except KeyError:
            raise KeyError(str(key_list[i]) + ' is missing')
            # break
def CheckOptionalKey(dict,optional_key_list):
	optional_key = []
	for i in range(len(optional_key_list)):
		try:
			dict[optional_key_list[i]]
		except KeyError:
			optional_key.append(optional_key_list[i])

	return optional_key
# config the parser to the right order:
def read_input_file(input_file,verbose=False):
    config = configparser.ConfigParser()
    config.read(input_file)
    config = config._sections

    # read into each dict
    file_min = ['Inputs','Populations','Mutations','Paths','Outputs']
    CheckKey(config,file_min)

    Inputs_dict = config['Inputs']
    Populations_dict = config['Populations']
    Mutations_dict = config['Mutations']
    Paths_dict = config['Paths']
    # Larch_dict = config['Larch_Paths']
    Outputs_dict = config['Outputs']

    # Checking for minimum inputs
    #Removed 'data_cutoff' -evan
    input_min = ['data_file','output_file']
    CheckKey(Inputs_dict,input_min)

    population_min = ['population','num_gen','best_sample','lucky_few']
    CheckKey(Populations_dict,population_min)

    mutation_min = ['chance_of_mutation','original_chance_of_mutation','mutated_options']
    CheckKey(Mutations_dict,mutation_min)

    path_min = ['npeaks']
    path_optional = ['path_optimize','steady_state','corr_paths']
    #print(Paths_dict.keys())
    CheckKey(Paths_dict,path_min)
    path_missing = CheckOptionalKey(Paths_dict,path_optional)


    # larch_min = ['kmin','kmax','kweight','deltak','rbkg','bkgkw','bkgkmax']
    # CheckKey(Larch_dict,larch_min)

    output_min =['print_graph','num_output_paths']
    output_optional = ['steady_state']
    CheckKey(Outputs_dict,output_min)
    output_missing = CheckOptionalKey(Outputs_dict,output_min)

    # Adjust values

    # Pack all of them into a single dicts
    file_dict = {}
    file_dict['Inputs'] = Inputs_dict
    file_dict['Populations'] = Populations_dict
    file_dict['Mutations'] = Mutations_dict
    file_dict['Paths'] = Paths_dict
    # file_dict['Larch_Paths'] = Larch_dict
    file_dict['Outputs'] = Outputs_dict

    if verbose == True:
        print_input_file(file_dict)

    return file_dict

def print_input_file(file_dict):
    for key,value in file_dict.items():
        print("[" +bcolors.BOLD + str(key)+ bcolors.ENDC +"]")
        for inner_key,inner_value in value.items():
            print('---'  +  inner_key + ": " +inner_value)


## Need to run some sample:
# if __name__ == '__main__':
#
#     checkKey()



"""
Created 7/5/23
File is a copy of xes_individual in the GA folder, but with some modifications to make it work with xes analysis file:
-Evan Restuccia (evan@restuccias.com)
"""
from tkinter import N

from matplotlib.bezier import get_parallels
from xes_neo.gui.xes_fit import peak,background

class Individual():
    """
    The Individual holds all peaks and backgrounds in the self.peakArr and self.bkgnArr
    (SVSC/peak shirley backgrounds are stored in each peak class)
    The arrays are populated individually and then used to manage and compile all factors external to each
    component of the fit.
    Generally, an individual is a fit in its entirety.

    get_params() will get all fit parameters
    mutate_() will mutate the entire fit
    """
    def __init__(self,backgrounds,peaks,scale_var,pars_range=''):
        """
        backgrounds (array) Array where each element is the name of the background type desired
        peaks (array) array where each element is the name of the desired peakType
        pars_range (dict) each key is the name of a parameter with a tuple that contains the range the parameter is allowed to explore

        """
        self.scale_var = scale_var
        self.SVSC_toggle = False
        self.pars_range = pars_range

        #both peaks and backgrounds are arrays of strings which represent the type of the background of the peak/bkgn
        self.nPeaks = len(peaks)
        self.nBackgrounds = len(backgrounds)
        self.peakArr = [None]* self.nPeaks
        self.bkgnArr = [None] * self.nBackgrounds

        """
        the Photon Energy needs to be personalized
        we take the range, which right now is something like (0,.2), i.e. the allowed variance in PE
        Then we add our photon energy to it to move the range to the right spot

        In the gui folder, we skip this try block, because we plan on custom setting these values later, and use a dummy paramRange of ''
        """
        if 'SVSC_shirley' in backgrounds:
            self.SVSC_toggle = True
            self.bkgnArr = [None] * (self.nBackgrounds-1)
        else:
            self.SVSC_toggle = False

        #print(self.peakArr)
        
        
        #I dont think we need this:
        if pars_range != '':
            pars_range['Photon Energy'][0],pars_range['Photon Energy'][1] = PE1,PE2
            pars_range['spinOrbitSplitting'][0],pars_range['spinOrbitSplitting'][1] = SOS1,SOS2
            pars_range['Gaussian'][0],pars_range['Gaussian'][1] = S1,S2
            pars_range['Lorentzian'][0],pars_range['Lorentzian'][1] = G1,G2
            pars_range['Amplitude'][0],pars_range['Amplitude'][1] = A1,A2
            pars_range['Lorentzian Coster-Kronig'][0],pars_range['Lorentzian Coster-Kronig'][1] = CK1,CK2
            

        
    

        for i in range(self.nPeaks):
            #self.peakArr[i] = peak(pars_range,peaks[i])
          
            self.peakArr[i] = peak(pars_range,peaks[i])

            if self.SVSC_toggle:
                self.peakArr[i].SVSC_toggle(True)
        '''
        j=0
        for i,bkgn in enumerate(backgrounds):
            #print(bkgn)
            if bkgn.lower() == 'svsc_shirley':
                #print("Pass")
                j+=1
                pass
            else:
                self.bkgnArr[i-j] = background(pars_range,bkgn, peaks[i])
        '''
        #each index in the peaks/background array is the name of the peak/background type to be used
        k=0
        n=0
        

        for i in range(self.nBackgrounds):
            if backgrounds[i] == 'SVSC_shirley':
                self.nBackgrounds -=1
                pass
            elif backgrounds[i] == 'Shirley-Sherwood':

                self.bkgnArr[k] = background(pars_range,backgrounds[i], peaks[n]) #Creating a unique shirley background for each peak...I think

                k +=1
                n +=1

            else:

                #n = 0 #Doesn't matter which peak we are looking at --> Only matters in Shirley case
                self.bkgnArr[k] = background(pars_range,backgrounds[i], peaks[n])

                k+=1




    def add_peak(self,peakType):
        self.peakArr.append(peak(self.pars_range,peakType))
       
           
    def add_bkgn(self,bkgnType):
        self.bkgnArr.append(background(self.pars_range,bkgnType))


    #adds all backgrounds and peaks as one y value array
    def getFit(self,x,y, peakType, backgrounds):
        
        yFit = [0]*len(x)

        for i in range(self.nPeaks):
           
            if(peakType == "Doniach-Sunjic"):
                asymD = self.peakArr[i].asymmetryDoniach
            else:
                asymD = 0
            PE = self.peakArr[i].photonEnergy
            width = self.peakArr[i].lorentz
            sigma = self.peakArr[i].gaussian
            A = self.peakArr[i].amp
            asym = self.peakArr[i].asymmetry
            singlet = self.peakArr[i].is_singlet
            coster_kronig = self.peakArr[i].is_coster_kronig #I do not know why but this is not taking in the correct bool statement for Coster-Kronig, only takes in init value in xes_fit gui
          
            if coster_kronig == True:
                width_CK = self.peakArr[i].lorentz_CK
            else:
                width_CK = self.peakArr[i].lorentz
            branch = self.peakArr[i].branching_ratio
            split = self.peakArr[i].spinOrbitSplit

            #asymD = self.peakArr[i].asymmetryDoniach
            if self.SVSC_toggle:
                peak_y,svsc_y = self.peakArr[i].getY(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)
                yFit += peak_y
                yFit += svsc_y
            else:
                yFit += self.peakArr[i].getY(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)

        n = 0
        for i in range(self.nBackgrounds):

           
            if(peakType == "Doniach-Sunjic"):
                asymD = self.peakArr[i].asymmetryDoniach
            else:
                asymD = 0
            PE = self.peakArr[n].photonEnergy
            width = self.peakArr[n].lorentz
            sigma = self.peakArr[n].gaussian
            A = self.peakArr[n].amp
            asym = self.peakArr[n].asymmetry
            singlet = self.peakArr[n].is_singlet
            coster_kronig = self.peakArr[n].is_coster_kronig
            if coster_kronig == True:
                width_CK = self.peakArr[n].lorentz_CK
            else:
                width_CK = self.peakArr[n].lorentz
            branch = self.peakArr[n].branching_ratio
            split = self.peakArr[n].spinOrbitSplit
           
            
            yFit += self.bkgnArr[i].getY(x,y, PE , width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split, self.scale_var) #ISSUE COMES FROM HERE --> NOT SEEING THE self.photonEnergy or self.lorentz VALUES

            if backgrounds[i] == 'Shirley-Sherwood':
                n += 1

        return yFit

    def getFWHM(self,x, peakType):
        FWHM = [0]*self.nPeaks
        for i in range(self.nPeaks):
            if(peakType == "Doniach-Sunjic"):
                asymD = self.peakArr[i].asymmetryDoniach
            else:
                asymD = 0
            PE = self.peakArr[i].photonEnergy
            width = self.peakArr[i].lorentz
            sigma = self.peakArr[i].gaussian
            A = self.peakArr[i].amp
            asym = self.peakArr[i].asymmetry
            singlet = self.peakArr[i].is_singlet
            coster_kronig = self.peakArr[i].is_coster_kronig
            if coster_kronig == True:
                width_CK = self.peakArr[i].lorentz_CK
            else:
                width_CK = self.peakArr[i].lorentz
            branch = self.peakArr[i].branching_ratio
            split = self.peakArr[i].spinOrbitSplit
            FWHM[i] = self.peakArr[i].getFWHM(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)

        return FWHM





    def getFitWithComponents(self,x,y, peakType, backgrounds):
        """
        gets fit and also returns all the components that make up the fit
        """
        yFit = [0]*len(x)
        #print(self.peakArr)
        bkgn_components_arr = []
        peak_components_arr = []
        for i in range(self.nPeaks):
            PE = self.peakArr[i].photonEnergy
            width = self.peakArr[i].lorentz
            sigma = self.peakArr[i].gaussian
            A = self.peakArr[i].amp
            asym = self.peakArr[i].asymmetry
            singlet = self.peakArr[i].is_singlet
            coster_kronig = self.peakArr[i].is_coster_kronig
            if coster_kronig == True:
                width_CK = self.peakArr[i].lorentz_CK
            else:
                width_CK = self.peakArr[i].lorentz
            branch = self.peakArr[i].branching_ratio
            split = self.peakArr[i].spinOrbitSplit

            if(peakType == "Doniach-Sunjic"):
                asymD = self.peakArr[i].asymmetryDoniach
            else:
                asymD = 0

            if self.SVSC_toggle:
                peakComp,svsc =  self.peakArr[i].getY(x)
                yFit += peakComp
                yFit += svsc
                bkgn_components_arr.append(svsc)
                peak_components_arr.append(peakComp)

            else:

                peakComp =  self.peakArr[i].getY(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)
                yFit += peakComp

                peak_components_arr.append(peakComp)

        n = 0
        total_bkgn = [0]*len(x)
        total_bkgn = [0]*len(x)
        for i in range(self.nBackgrounds):


            if(peakType == "Doniach-Sunjic"):
                asymD = self.peakArr[i].asymmetryDoniach
            else:
                asymD = 0
            PE = self.peakArr[n].photonEnergy
            width = self.peakArr[n].lorentz
            sigma = self.peakArr[n].gaussian
            A = self.peakArr[n].amp
            asym = self.peakArr[n].asymmetry
            singlet = self.peakArr[n].is_singlet
            coster_kronig = self.peakArr[n].is_coster_kronig
            if coster_kronig == True:
                width_CK = self.peakArr[n].lorentz_CK
            else:
                width_CK = self.peakArr[n].lorentz
            branch = self.peakArr[n].branching_ratio
            split = self.peakArr[n].spinOrbitSplit
            bkgnComp = self.bkgnArr[i].getY(x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split,self.scale_var)

            yFit += bkgnComp
            bkgn_components_arr.append(bkgnComp)
            if backgrounds[i] == 'Shirley-Sherwood':
                n += 1

        #print("BACKGROUND SUM =", net_background)
                n += 1

        #print("BACKGROUND SUM =", net_background)
        #inefficient but only needs to work a few times so it should be fine
        for i in range(len(peak_components_arr)):
            for l in range(len(bkgn_components_arr)):
                for k in range(len(x)):
                    peak_components_arr[i][k] += bkgn_components_arr[l][k]


        for i in range(len(bkgn_components_arr)):
            for j in range(len(y)):
                total_bkgn[j] += bkgn_components_arr[i][j]
        net_bkgn_left = []
        net_bkgn_right = []
        for i in range(30):
            net_bkgn_left.append(total_bkgn[i] - y[i])

        for i in range(len(y)-1, len(y)-30, -1):
            net_bkgn_right.append(total_bkgn[i] - y[i])
        #Create some kind of check to see if this value is negative or if its positive value is above %*average_left_&_right of y
        #print("NET BACKGROUND DIFFERENCE LEFT", sum(net_bkgn_left)/len(net_bkgn_left))
        #print("NET BACKGROUND DIFFERENCE RIGHT", sum(net_bkgn_right)/len(net_bkgn_right))





        for i in range(len(bkgn_components_arr)):
            for j in range(len(y)):
                total_bkgn[j] += bkgn_components_arr[i][j]
        net_bkgn_left = []
        net_bkgn_right = []
        for i in range(30):
            net_bkgn_left.append(total_bkgn[i] - y[i])

        for i in range(len(y)-1, len(y)-30, -1):
            net_bkgn_right.append(total_bkgn[i] - y[i])
        #Create some kind of check to see if this value is negative or if its positive value is above %*average_left_&_right of y
        #print("NET BACKGROUND DIFFERENCE LEFT", sum(net_bkgn_left)/len(net_bkgn_left))
        #print("NET BACKGROUND DIFFERENCE RIGHT", sum(net_bkgn_right)/len(net_bkgn_right))



        return yFit,peak_components_arr,bkgn_components_arr

    def get(self):
        """
        Get the whole set
        """
        
        return (self.peakArr + self.bkgnArr)

    def get_params(self,is_coster_kronig):
        params = []
       
        
        for i in range(len(self.peakArr)):
            params.append(self.peakArr[i].get(is_coster_kronig[i]))
        
        #print(self.bkgnArr[1].get(), "HELLO")
        #print(self.bkgnArr[1].get(), "HELLO")
        for i in range(len(self.bkgnArr)):
            params.append(self.bkgnArr[i].get())
        return params

    def get_peak(self,i):
        return self.peakArr[i].get()

    def get_peaks(self):
        return self.peakArr

    def get_background(self,i):
        return self.bkgnArr[i]
    def get_backgrounds(self):
        return self.bkgnArr

    def mutate_(self,chance):
        for peak in self.peakArr:
            peak.mutate(chance)
        for bkgn in self.bkgnArr:
            bkgn.mutate(chance)


    def verbose(self):
        """
        Print out the Populations
        """
        for i in range(self.npaths):
            self.Population[i].verbose()

    ''' Could be useful later I just dont want to write it
    def set_peak(self,i,A,h_f,m):
        self.Population[i].set(A,h_f,m)
    '''



import tkinter as tk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
import EXAFS_Analysis
import os


class Analysis_Plot:
    def __init__(self,frame,larch):
        self.mylarch = larch
        self.fig = Figure(figsize=(3.5,3.5), dpi=100)
        self.canvas = FigureCanvasTkAgg(self.fig, master=frame)
        # Create inital figure canvas
        self.canvas.get_tk_widget().grid(column=0,row=10,columnspan=4,sticky="nsew",
            padx = 5,pady=5)
        self.ax = self.fig.add_subplot(111)
        # create toolbar
        self.toolbarFrame = tk.Frame(master=frame)
        self.toolbarFrame.grid(column=0,row=11,columnspan=4,sticky="nsew")
        toolbar = NavigationToolbar2Tk(self.canvas, self.toolbarFrame)
        self.params = {}

    def setup_params(self,params):
        self.params = params

    def setup_dirs(self,dir):
        self.dir = dir

    def setup_paths(self,paths):
        self.paths = paths

    def extract_and_run(self,output_folder):

        self.EXAFS_Analysis = EXAFS_Analysis.EXAFS_Analysis(self.paths,self.dir,self.params)
        self.EXAFS_Analysis.extract_data()

        self.EXAFS_Analysis.larch_init()
        self.EXAFS_Analysis.larch_score()
        self.EXAFS_Analysis.individual_fit()
        self.EXAFS_Analysis.construct_latex_table()

        self.EXAFS_Analysis.export_files(header='Test',dirs=output_folder.get())
        # Cu_10.export_igor_individual('export/Cu_10/Cu_10.ipf')
        self.EXAFS_Analysis.export_igor_individual(os.path.join(output_folder.get(),'Test.ipf'))

        return self.EXAFS_Analysis.return_str,self.EXAFS_Analysis.latex_table_str

    def plot_k_r_space(self):
        self.fig.clf()
        self.EXAFS_Analysis.plot(fig_gui=self.fig)
        self.canvas.draw()

    def plot_individual(self):
        self.fig.clf()
        self.EXAFS_Analysis.individual_fit(fig_gui=self.fig)
        self.canvas.draw()
        
    def plot_error(self):
        self.fig.clf()
        self.EXAFS_Analysis.plot_error(fig_gui=self.fig)
        self.canvas.draw()

    def plot_occurances(self):
        self.fig.clf()
        self.EXAFS_Analysis.plot_occ_list(fig_gui=self.fig)
        self.canvas.draw()



"""
Author: Evan Restuccia (evan@restuccias.com)
"""

from turtle import back
from matplotlib import get_backend
import numpy as np
#from pyparsing import None_debug_action
import scipy as scipy
from scipy import signal
import random
import decimal
from numba.core.errors import NumbaDeprecationWarning, NumbaPendingDeprecationWarning #Use this to get rid of numba warnings that are outputted in the terminal during run
import warnings
import numba as nb


warnings.filterwarnings('ignore')
warnings.simplefilter('ignore', category=NumbaDeprecationWarning)
warnings.simplefilter('ignore', category=NumbaPendingDeprecationWarning)
#import chart_studio.plotly as py
#import peakutils
#from pybaselines import Baseline, utils #Will have to have people install this in order to make it run
#This is a test
class peak():
    #Probably default these to -1 later as a test condition, but not yet sure
    def __init__(self,paramRange,peakType, is_singlet = True, is_coster_kronig = False): #Coster-Kronig only works rn if this is True, but it will be true for all peaks.
        
        """
        Skips  the try catch block here, because the analysis which requires the fit
        will custom set the peaks later with values read from the GA output

        To add a Peak type:
        Add its peakType option to the picker in the GUI
        Give its parameters ranges and pass them into the write_ini in the gui
        add the range to ini_parser and also the paramrange dict in xes_neo
        add that bkgnType in init here, and assign the function to self.peakFunc
        add the get function for it
        add the set function in the gui xes_fit
        """
        
        #fetch ranges for the values from the dict
        self.SVSC = False
        self.paramRange= paramRange
        self.is_singlet = is_singlet
        self.is_coster_kronig = is_coster_kronig
       
        if paramRange != '':
            self.gaussRange = np.arange(paramRange['Gaussian'][0],paramRange['Gaussian'][1],paramRange['Gaussian'][2])
            self.lorentzRange = np.arange(paramRange['Lorentzian'][0],paramRange['Lorentzian'][1],paramRange['Lorentzian'][2])

            self.lorentzCKRange = np.arange(paramRange['Lorentzian Coster-Kronig'][0],paramRange['Lorentzian Coster-Kronig'][1],paramRange['Lorentzian Coster-Kronig'][2])

            self.photonEnergyRange = np.arange(paramRange['Photon Energy'][0],paramRange['Photon Energy'][1],paramRange['Photon Energy'][2])
            self.ampRange = np.arange(paramRange['Amplitude'][0],paramRange['Amplitude'][1],paramRange['Amplitude'][2])
            self.asymmetryRange = np.arange(paramRange['Asymmetry'][0],paramRange['Asymmetry'][1],paramRange['Asymmetry'][2])
            self.asymmetryDoniachRange = np.arange(paramRange['Asymmetry Doniach-Sunjic'][0],paramRange['Asymmetry Doniach-Sunjic'][1],paramRange['Asymmetry Doniach-Sunjic'][2])
            
            
            
            try:
                self.spinOrbitSplitRange = np.arange(paramRange['spinOrbitSplitting'][0],paramRange['spinOrbitSplitting'][1],paramRange['spinOrbitSplitting'][2])
            except:
                self.spinOrbitSplitRange = [0,0,0]


            #fully free within their range
            self.gaussian = np.random.choice(self.gaussRange)
            self.lorentz = np.random.choice(self.lorentzRange)
            #self.lorentz_CK = np.random.choice(self.lorentzCKRange) #Coster-Kronig
            #self.lorentz_CK = self.lorentz/1
            self.lorentz_CK = np.random.choice(self.lorentzCKRange)
            self.amp = np.random.choice(self.ampRange) #Scaling issues --> figure out what is wrong
            self.asymmetry = np.random.choice(self.asymmetryRange)
            self.asymmetryDoniach = np.random.choice(self.asymmetryDoniachRange)
            


            #the range is a modifier on the input value
            self.photonEnergy= np.random.choice(self.photonEnergyRange) #Not using rn --> changed to set parameter
            #self.s_o_split = np.random.choice(self.s_o_splittingRange) #Not using rn
        else:
            self.gaussian = None
            self.lorentz = None
            self.lorentz_CK = None
            self.photonEnergy = None
            self.amp = None
            self.asymmetry = None
        self.peakType = peakType
        self.peak_y = []

        
        if paramRange != '':
            self.branching_ratio = paramRange['branching_ratio']
            
            self.spinOrbitSplit= np.random.choice(self.spinOrbitSplitRange)
        else:
            self.branching_ratio = 1
            self.spinOrbitSplit = None
        
        
        self.SVSC = False
        self.peakType = peakType
        if(self.peakType.lower() == "voigt"):
            self.func = self.voigtFunc
        elif(self.peakType.lower() == "gaussian"):
            self.func = self.gaussFunc
        elif(self.peakType.lower() == "lorentzian"):
            self.func = self.lorentzFunc
        elif(self.peakType.lower() == "double lorentzian"):
            self.func = self.doubleLorentzFunc
        elif(self.peakType.lower() == "doniach-sunjic"):
            self.func = self.doniachSunjicFunc

        else:
            print("Error assigning peak type")
            print("Peaktype found is: " + str(self.peakType))
            exit()
        


    def peakFunc(self,x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        return self.func(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)

    #----------------------------Getters-------------------------------------------------------

    def get(self,is_coster_kronig,for_output_data_file = False):
        
        """
        Gets params in format [Singlet Params, bool for is_singlet, doublet params, peakType]
        """
        #Add in FWHM values for each curve type --> Voigt has 3, gauss and loretnz only have one
        params = []
        if self.peakType.lower() == 'voigt':
            params = [self.photonEnergy,self.gaussian,self.lorentz,self.amp] #mutate relies on the order here, so to change this you need to change
            #params = [self.photonEnergy,self.gaussian,self.lorentz,self.amp,self.peakType] #mutate relies on the order here, so to change this you need to change mutate
        if self.peakType.lower() == 'gaussian':
            params = [self.photonEnergy,self.gaussian,self.amp]#everything except lorentzian which in this case is the width i believe --> Still want FWHM
            #params = [self.photonEnergy,self.gaussian,self.amp,self.peakType]#everything except lorentzian which in this case is the width i believe --> Still want FWHM
        if self.peakType.lower() == 'lorentzian':
            params = [self.photonEnergy,self.lorentz,self.amp]
            #params = [self.photonEnergy,self.lorentz,self.amp,self.peakType]
        if self.peakType.lower() == 'double lorentzian':
            params = [self.photonEnergy,self.gaussian,self.lorentz,self.amp,self.asymmetry]
            #params = [self.photonEnergy,self.gaussian,self.lorentz,self.amp,self.asymmetry,self.peakType]
        if self.peakType.lower() == 'doniach-sunjic':
            params = [self.photonEnergy,self.gaussian,self.lorentz,self.amp,self.asymmetry] #no gaussian needed
            #params = [self.photonEnergy,self.lorentz,self.amp,self.asymmetry,self.peakType] #no gaussian needed
        
        #grab the bool is_singlet if writing for output
        if(for_output_data_file):
            params.append(self.is_singlet)
        self.is_coster_kronig = is_coster_kronig
      
    

       
        
       
            
            
        #ISSUE: INDIVIDUAL NOT SEEING COSTER KRONIG AS TRUE UNLIKE SINGLET. NOT SURE WHY???


        #grab the doublet params if its a doublet
       
        if not(self.is_singlet):
            if(for_output_data_file):
                params.append(self.branching_ratio)
            params.append(self.spinOrbitSplit)

            if self.is_coster_kronig: #If it is a doublet AND is Coster-Kronig
              
                params.append(self.lorentz_CK) #This is not being seen right now therefore not being outputted in analysis tab. Change if self.is_coster_kronig to if not()? why?
                #params.append(self.is_coster_kronig)
            params.append(self.is_coster_kronig)
            
          


            

        #and always end on the peakType
        params.append(self.peakType)



        #Add this peaks shirley background if using peak shirley
        if self.SVSC:
            SVSC_params = self.SVSC_background.get()
            for param in SVSC_params:
                params.append(param)
            return params
        else:
            if len(params) == 0:
                print("Cant do 'def get' in peaks class in XES_FIT, most likely a new peak was added and needs to be added to the get options")
                exit()
            else:
                return params

    """
    I don't think this set of functions is used ever
    """
    def getGaussian(self):
        return self.gaussian
    def getLorenztian(self):
        return self.lorentz
    def getAmplitude(self):
        return self.amp
    def getPhotonEnergy(self):
        return self.photonEnergy
    def getAsymmetry(self):
        return self.asymmetry
    def getAsymmetryDoniach(self):
        return self.asymmetryDoniach
    def getSpinOrbitSplit(self):
        return self.spinOrbitSplit



    def getFWHM(self,x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split): #Not wure where to call this in order to output the FWHM values for each peak curve type
        self.photonEnergy = PE
        self.lorentz = width
        self.lorentz_CK = width_CK
        self.is_coster_kronig = coster_kronig
        self.gaussian = sigma
        self.amp = A
        self.asymmetry = asym
        self.asymmetryDoniach = asymD
        self.is_singlet = singlet
        self.branching_ratio = branch
        self.spinOrbitSplit = split

        #MAD = np.mean(np.absolute(self.peak_y-np.mean(self.y))) #Mean Absolute Deviaition
        #self.lorentz = gamma = FWHM of a pure Lorentzian lineshape
        if(self.peakType.lower() == "voigt"):

            self.fwhm_g = 2*self.gaussian*np.sqrt(2*np.log(2))
            self.fwhm_l = self.lorentz
            self.fwhm_v = 0.5346*self.fwhm_l + np.sqrt(0.2166*pow(self.fwhm_l,2) + pow(self.fwhm_g, 2))
            #self.fwhm_v = self.fwhm_l/2 + np.sqrt(pow(self.fwhm_l, 2)/4 + pow(self.fwhm_g, 2))

            return self.fwhm_v #, self.fwhm_l, self.fwhm_g
        elif(self.peakType.lower() == "gaussian"):
            self.fwhm_g = 2*self.gaussian*np.sqrt(2*np.log(2))
            #print(self.fwhm_g)
            return self.fwhm_g
        elif(self.peakType.lower() == "lorentzian"):
            self.fwhm_l = self.lorentz
            return self.fwhm_l
        elif(self.peakType.lower() == "double lorentzian"):
            self.fwhm_g = 2*self.gaussian*np.sqrt(2*np.log(2))
            self.fwhm_l = self.lorentz
            self.fwhm_v = 0.5346*self.fwhm_l + np.sqrt(0.2166*pow(self.fwhm_l,2) + pow(self.fwhm_g, 2))
            #self.fhwm_v = self.fwhm_l/2 + np.sqrt(pow(self.fwhm_l, 2)/4 + pow(self.fwhm_g, 2))
            return self.fwhm_v #, self.fwhm_l, self.fwhm_g #Should double lorentzian still provide FWHM_L?
        elif(self.peakType.lower() == "doniach-sunjic"):
            self.fwhm_g = 2*self.gaussian*np.sqrt(2*np.log(2))
            self.fwhm_l = self.lorentz #Should this be two values? One for the asymmetrical side of the peak
            self.fwhm_v = 0.5346*self.fwhm_l + np.sqrt(0.2166*pow(self.fwhm_l,2) + pow(self.fwhm_g, 2))
            #self.fhwm_v = self.fwhm_l/2 + np.sqrt(pow(self.fwhm_l, 2)/4 + pow(self.fwhm_g, 2))
            return self.fwhm_v #, self.fwhm_l, self.fwhm_g #Should double lorentzian still provide FWHM_G?
        else:
            print("Error in FWHM caluclation")
            print("Peaktype found is: " + str(self.peakType))
            pass
    def getY(self,x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        """
        Gets the y values for a peak by running the peakFunc, then, if it has an SVSC background, it will tell its
        SVSC background to run and return both
        """
        #self.FWHM_values = self.getFWHM()
        #print("FWHM: ",self.FWHM_values)
    
        self.peakFunc(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)
        if self.SVSC:
            SVSC_vals = self.SVSC_background.getY(x,self.peak_y)
            return self.peak_y,SVSC_vals
        return self.peak_y

    #------------------------SVSC "On" Switch and SVSC handling----------------------------

    def SVSC_toggle(self,boolVal):
        """
        Turns on the SVSC for a peak if the option is turned on and creates a peak for it
        """
        self.SVSC = boolVal
        self.SVSC_background = background(self.paramRange,'SVSC_shirley')

    def set_svsc_shirley(self,val):
        self.SVSC_background.set_shirley_sherwood(val)

     #------------------------Setters--------------------------------------

    def set(self,params,doublet_params, CK_params, is_singlet=False, is_coster_kronig=True): #Changed is_coster_kronig to True here. WHy is is_singlet set to False here?
        """
        This is the main differentiator between the fit in gui and xesfolders
        The set function takes in a list of params, and should expect them to come in the same order that they were pushed out
        in the get function.

        """
        if(self.peakType.lower() == "voigt"):
            self.set_voigt(params)
        elif(self.peakType.lower() == "gaussian"):
            self.set_gauss(params)
        elif(self.peakType.lower() == "lorentzian"):
            self.set_lorentz(params)
        elif(self.peakType.lower() == "double lorentzian"):
            self.set_doubleLorentz(params)
        elif(self.peakType.lower() == "doniach-sunjic"):
            self.set_doniachSunjic(params)
        else:
            print("Error, cant find type to set, type found was: " + str(self.peakType))
  
        if not is_singlet:
            self.is_singlet = is_singlet
            self.branching_ratio = doublet_params[0]
            self.spinOrbitSplit = doublet_params[1]
          
            if is_coster_kronig == True:

                self.lorentz_CK = CK_params[0]


    #probably can delete these functions
    def setGaussian(self,newVal):
        self.gaussian = newVal
    def setLorentzian(self,newVal):
        self.lorentz = newVal
    def setAmplitude(self,newVal):
        self.amp = newVal
    def setPhotonEnergy(self,newVal):
        self.photonEnergy = newVal
    def setAsymmetry(self,newVal):
        self.asymmetry = newVal
    def setAsymmetryDoniach(self,newVal):
        self.asymmetryDoniach = newVal
    def setSpinOrbitSplit(self,newVal):
        self.spinOrbitSplit = newVal





    def set_voigt(self,paramList):
        self.photonEnergy = paramList[0]
        self.gaussian = paramList[1]
        self.lorentz = paramList[2]
        self.amp = paramList[3]
        #self.FWHM_values = paramList[4]

    def set_gauss(self,paramList):
        self.photonEnergy = paramList[0]
        self.gaussian = paramList[1]
        self.amp = paramList[2]
        #self.FWHM_values = paramList[3]
        #does lorentz effect gauss? --> it is just the width which is not used in the equation at all

    def set_lorentz(self,paramList):
        self.photonEnergy = paramList[0]
        self.lorentz = paramList[1]
        self.amp = paramList[2]
        #self.FWHM_values = paramList[3]

    def set_doubleLorentz(self,paramList):
        self.photonEnergy = paramList[0]
        self.gaussian = paramList[1]
        self.lorentz = paramList[2]
        self.amp = paramList[3]
        self.asymmetry = paramList[4]
        #self.FWHM_values = paramList[5]

    def set_doniachSunjic(self,paramList):
        self.photonEnergy = paramList[0]
        self.gaussian = paramList[1]
        self.lorentz = paramList[2]
        self.amp = paramList[3]
        self.asymmetryDoniach = paramList[4]
        #self.FWHM_values = paramList[5]

    #---------------------------------------Mutation Functions-------------------------------------------------


    def mutate(self,chance):
        self.mutateGauss(chance)
        self.mutateAmplitude(chance)
        self.mutatePE(chance)
        self.mutateLorentz(chance)
        self.mutateAsymmetry(chance)
        self.mutateAsymmetryDoniach(chance)
        '''
        if is_singlet==False:
            self.mutateSpinOrbitSplit
        '''
        if(self.SVSC):
            self.SVSC_background.mutate(chance)

    def mutateGauss(self,chance):
        if random.random()*100 < chance:
            self.gaussian = np.random.choice(self.gaussRange)
    def mutateLorentz(self,chance):
        if random.random()*100 < chance:
            self.lorentz = np.random.choice(self.lorentzRange)
    def mutateAmplitude(self,chance):
        if random.random()*100 < chance:
            self.amp = np.random.choice(self.ampRange)
    def mutatePE(self,chance):
        if random.random()*100 < chance:
            self.photonEnergy = np.random.choice(self.photonEnergyRange)
    def mutateAsymmetry(self,chance):
        if random.random()*100 < chance:
            self.asymmetry = np.random.choice(self.asymmetryRange)
    def mutateAsymmetryDoniach(self,chance):
        if random.random()*100 < chance:
            self.asymmetryDoniach = np.random.choice(self.asymmetryDoniachRange)
    '''
    def mutateSpinOrbitSplit(self,chance):
        if random.random()*100 < chance:
            self.spinOrbitSplit = np.random.choice(self.spinOrbitSplitRange)
    '''

    #Peak curve fit equations start at line 6681 in Aanalyzer PUnit1


    #----------------------Peak Curve Form Definitions------------------------------------------#


    #A bit scrappy at the moment, may need cleaning later
    def voigtFunc(self,x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):


        #Check to see if the data is in KE energy
        self.photonEnergy = PE
        self.lorentz = width
        self.lorentz_CK = width_CK
        self.is_coster_kronig = coster_kronig
        self.gaussian = sigma
        self.amp = A
        self.asymmetry = asym
        self.asymmetryDoniach = asymD
        self.is_singlet = singlet
        self.branching_ratio = branch
        self.spinOrbitSplit = split

        """
        Calculate the Voigt lineshape with variable peak position and intensity using a convolution of a Gaussian and a Lorentzian distribution.

        Parameters:
            x (array-like): The input array of independent variables.
            self.photonEnergy (float): The position of the peak.
            self.gaussian (float): The standard deviation of the Gaussian distribution.
            self.lorentz (float): The full width at half maximum (FWHM) of the Lorentzian distribution.
            intensity (float): The intensity of the peak.

        Returns:
            array-like: The values of the Voigt lineshape at the given x-values.
        """
        if self.gaussian ==0:
            self.gaussian += self.paramRange['Gaussian'][2]


        data_range= max(x) - min(x)
        data_range /= 2
        middle = min(x)+data_range
        offset = self.photonEnergy-middle
        num_points = len(x)
        x_values, dx = np.linspace(-data_range,data_range,num_points,retstep=True)


        gaussian = np.exp(-np.power(x_values, 2) / (2*(np.power(self.gaussian, 2)))) / (self.gaussian * np.sqrt(2 * np.pi))

        if self.is_singlet == True:
            #lorentz
            #                        gamma
            #            ----------------------------
            #                           2    gamma   2
            #            2 Pi ((x - PE ) + ( -----  )  )
            #                                  2
            lorentzian = (self.lorentz / (2*np.pi*(np.power(x_values+offset, 2) + np.power(self.lorentz/2, 2))))

            #gaussian
            #                    2
            #        - (x - PE)
            #        ------------
            #                2
            #         2 sigma
            #    e
            #    -----------------
            #            _____
            #    sigma  |/2 Pi

            
            final_voigt = scipy.signal.convolve(gaussian,lorentzian,'same')

     

        if self.is_singlet == False: #If it is a doublet

            if self.is_coster_kronig == True:

                nu1 = self.lorentz
                de11 = np.power((x_values+offset), 2) + np.power((self.lorentz/2), 2)
                de12 = 2*np.pi*de11
                lorentzian1 = (nu1/de12)/(1-self.branching_ratio)

                nu2 = self.branching_ratio*self.lorentz_CK
                de21 = np.power(x_values+offset+self.spinOrbitSplit,2) + np.power(self.lorentz_CK/2,2)
                de22 = 2*np.pi*de21
                lorentzian2 = (nu2/de22)/(1-self.branching_ratio)
                lorentzian = lorentzian1 + lorentzian2

                
                #Attempted multiiple things but the doublet is still not working.
                final_voigt = scipy.signal.fftconvolve(gaussian,lorentzian,'same')
            else:

                nu1 = self.lorentz
                de11 = np.power((x_values+offset), 2) + np.power((self.lorentz/2), 2)
                de12 = 2*np.pi*de11
                lorentzian1 = (nu1/de12)/(1-self.branching_ratio)

                nu2 = self.branching_ratio*self.lorentz
                de21 = np.power(x_values+offset+self.spinOrbitSplit,2) + np.power(self.lorentz/2,2)
                de22 = 2*np.pi*de21
                lorentzian2 = (nu2/de22)/(1-self.branching_ratio)
                lorentzian = lorentzian1 + lorentzian2


             
                #Attempted multiiple things but the doublet is still not working.
                final_voigt = scipy.signal.fftconvolve(gaussian,lorentzian,'same')



        #normalize the height so that intensity is the height of the max of the peak

        scale = max(final_voigt)
        for i in range(len(final_voigt)):
            final_voigt[i] *= (self.amp/scale)

        #returns, but also updates the yValues of the fit to improve efficiency, we can call that instead of recalculating every time
        if x[0] < x[-1]:
            final_voigt = final_voigt[::-1]
        else:
            pass

        self.peak_y = final_voigt
        #peak.voigt = voigt

        return final_voigt



    def gaussFunc(self,x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        self.photonEnergy = PE
        self.lorentz = width
        self.lorentz_CK = width_CK
        self.is_coster_kronig = coster_kronig
        self.gaussian = sigma
        self.amp = A
        self.asymmetry = asym
        self.asymmetryDoniach = asymD
        self.is_singlet = singlet
        self.branching_ratio = branch
        self.spinOrbitSplit = split

        

        if self.gaussian == 0:
            self.gaussian = .01

        gaussian = np.exp(-np.power(x - self.photonEnergy, 2) / (2 * np.power(self.gaussian, 2))) / (self.gaussian * np.sqrt(2 * np.pi))

        gaussian_max = max(gaussian)
        if(gaussian_max != 0):
            peakAmp = self.amp/max(gaussian)
        else:
            peakAmp = 0
        gauss_curve = gaussian * peakAmp

        #self.peak_y = gauss_curve

        if self.is_singlet == False: #If it is a doublet
            gaussian1 = (np.exp(-np.power(x - self.photonEnergy, 2) / (2 * np.power(self.gaussian, 2))) / (self.gaussian * np.sqrt(2 * np.pi)))/(1-self.branching_ratio)
            gaussian2 = self.branching_ratio * (np.exp(-np.power((x - (self.photonEnergy + self.spinOrbitSplit)), 2) / (2 * np.power(self.gaussian, 2))) / (self.gaussian * np.sqrt(2 * np.pi)))/(1-self.branching_ratio)
            gaussian = gaussian1 + gaussian2

            gaussian_max = max(gaussian)
            if(gaussian_max != 0):
                peakAmp = self.amp/max(gaussian)
            else:
                peakAmp = 0
            gauss_curve = gaussian * peakAmp
        '''
        if x[0] < x[-1]: #KE CONDITION
            gauss_curve = gauss_curve[::-1]
        else:
            pass
        '''
        self.peak_y = gauss_curve
        
        return gauss_curve





    def lorentzFunc(self,x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):

        global photonEnergy
        self.photonEnergy = PE
        self.lorentz = width
        self.lorentz_CK = width_CK
        self.is_coster_kronig = coster_kronig
        self.gaussian = sigma
        self.amp = A
        self.asymmetry = asym
        self.asymmetryDoniach = asymD
        self.is_singlet = singlet
        self.branching_ratio = branch
        self.spinOrbitSplit = split


        data_range= max(x) - min(x)
        data_range /= 2

        middle = min(x)+data_range
        offset = self.photonEnergy-middle
        num_points = len(x)
        x_values, dx = np.linspace(-data_range,data_range,num_points,retstep=True)

      
        #self.peak_y = lorentzian
        if self.is_singlet == True:
            lorentzian = (self.lorentz / (2*np.pi*(np.power(x_values+offset, 2) + np.power(self.lorentz/2, 2))))
            scale = max(lorentzian)
            for i in range(len(lorentzian)):
                lorentzian[i] *= (self.amp/scale)




        elif self.is_singlet == False: #If it is a doublet
            #No clue why but self.lorentz does not need to be made into the HWHM for the doublet. Branching ratio is still off for some reason.
            if self.is_coster_kronig == True:
                nu1 = self.lorentz
                de11 = np.power((x_values+offset), 2) + np.power((self.lorentz/2), 2)
                de12 = 2*np.pi*de11
                lorentzian1 = (nu1/de12)/(1-self.branching_ratio)

                nu2 = self.branching_ratio*self.lorentz_CK
                de21 = np.power(x_values+offset+self.spinOrbitSplit,2) + np.power(self.lorentz_CK/2,2)
                de22 = 2*np.pi*de21
                lorentzian2 = (nu2/de22)/(1-self.branching_ratio)

                lorentzian = (lorentzian1 + lorentzian2)


                scale = max(lorentzian)
                for i in range(len(lorentzian)):
                    lorentzian[i] *= (self.amp/scale)

            else:

                nu1 = self.lorentz
                de11 = np.power((x_values+offset), 2) + np.power((self.lorentz/2), 2)
                de12 = 2*np.pi*de11
                lorentzian1 = (nu1/de12)/(1-self.branching_ratio)

                nu2 = self.branching_ratio*self.lorentz
                de21 = np.power(x_values+offset+self.spinOrbitSplit,2) + np.power(self.lorentz/2,2)
                de22 = 2*np.pi*de21
                lorentzian2 = (nu2/de22)/(1-self.branching_ratio)

                lorentzian = (lorentzian1 + lorentzian2)


                scale = max(lorentzian)
                for i in range(len(lorentzian)):
                    lorentzian[i] *= (self.amp/scale)

        

        if x[0] < x[-1]: #KE CONDITION
            lorentzian = lorentzian[::-1]
        else:
            pass

        self.peak_y = lorentzian


        return lorentzian






    def doubleLorentzFunc(self,x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        #Same as voigt but with an added asymmetry factor to the width (lorentz)in the lorentzian equation
        self.photonEnergy = PE
        self.lorentz = width
        self.lorentz_CK = width_CK
        self.is_coster_kronig = coster_kronig
        self.gaussian = sigma
        self.amp = A
        self.asymmetry = asym
        self.asymmetryDoniach = asymD
        self.is_singlet = singlet
        self.branching_ratio = branch
        self.spinOrbitSplit = split


        #print("LORENTZ CK b4", self.lorentz_CK)
        #self.lorentz_CK = self.lorentz/self.branching_ratio
       
        if self.gaussian == 0:
            self.gaussian = .01

        data_range= max(x) - min(x)
        data_range /= 2

        middle = min(x)+data_range
        #offset = self.photonEnergy-middle
        #sos_offset = offset - self.spinOrbitSplit//2 #offset for spin-orbit splitting
        #sos_offset = -sos_offset
        #print(offset, sos_offset)

        num_points = len(x)
        x_values, dx = np.linspace(-data_range,data_range,num_points,retstep=True)
        dx = round(dx, 5)

        mult = 1
        #common multipliers
        offset_adj = 1
        #common multipliers
        if dx >= 0.1:
            offset_adj = 0.3
        else:
            offset_adj = 0.35
        self.photonEnergy = self.photonEnergy #+ dx*mult
        offset = self.photonEnergy-middle - offset_adj #Data is offset by dx*val for some reason
      
      
        #gaussian = np.exp(-np.power(x - self.photonEnergy, 2) / (2 * np.power(self.gaussian, 2))) / (self.gaussian * np.sqrt(2 * np.pi))
        gaussian = np.exp(-np.power(x_values, 2) / ((np.power(self.gaussian, 2)))) / (self.gaussian * np.sqrt(np.pi)) #Took away 2*np.power(self.gaussian,2) and np.sqrt(2*np.pi)

        # Calculate the Lorentzian component

        #Added +offset to all Lorentzian functions instead of Gauss --> It has to be inside the equation to make the leftside of the peak more lorentzian
        new_x = np.zeros(len(x))
        numP = len(x)
        HWHM = self.lorentz/2
        lorentzLeft = HWHM*self.asymmetry #Width of left side of peak due to asymmetry

        HWHM_CK = self.lorentz_CK/2
        lorentzLeft_CK = HWHM_CK*self.asymmetry
        #print("LORENTZ CK", self.lorentz_CK)
       
        #z = np.arange(-xRange, xRange+stepSize,.05)
        yDoubleL = [0]*numP
        y_left = [0]*numP
        y_right = [0]*numP
        #doubleL_lower = [0]*numP #lower PE doublet peak
        #doubleL_higher = [0]*numP #higher PE doublet peak


        #For some reason the peak location gets offset with difference asymmetry values
        #Dependent on dx value......
        asym_offset = 0
        if self.asymmetry < 1.9:
            asym_offset = 0.4
        elif 1.9 <= self.asymmetry < 2.7:
            asym_offset = 0.8
        elif 2.7 <= self.asymmetry < 3.7:
            asym_offset = 1.2
        elif 3.7 <= self.asymmetry < 4.9:
            asym_offset = 1.6
        elif 4.9 <= self.asymmetry < 6.5:
            asym_offset = 2.0
        elif 6.5 <= self.asymmetry < 8.6:
            asym_offset = 2.4
        elif 8.6 <= self.asymmetry < 11.2:
            asym_offset = 2.8
        elif 11.2 <= self.asymmetry < 14.6:
            asym_offset = 3.2
        elif 14.6 <= self.asymmetry < 19.1:
            asym_offset = 3.6
        elif self.asymmetry >= 19.1:
            asym_offset = 4.0


        





        if self.is_singlet == True:
            for i in range(len(x)):

                #KE Condition
                if x[0] < x[-1]:
                    if round(x_values[i] - offset-asym_offset, 5) <= 0:

                        yDoubleL[i] = 1 / ( 1 + np.power( (x_values[i] - offset-asym_offset)/lorentzLeft, 2 ) ) / np.pi
                    else:

                        yDoubleL[i] = 1 / ( 1 + np.power( (x_values[i] - offset-asym_offset)/HWHM, 2 ) ) / np.pi
                    yDoubleL[i] = yDoubleL[i] / (lorentzLeft/2 + HWHM/2)


                else:
                    if round(x_values[i] + offset, 5) >= 0:

                        yDoubleL[i] = 1 / ( 1 + np.power( (x_values[i] + offset)/lorentzLeft, 2 ) ) / np.pi
                    else:

                        yDoubleL[i] = 1 / ( 1 + np.power( (x_values[i] + offset)/HWHM, 2 ) ) / np.pi

                yDoubleL[i] = yDoubleL[i] / (lorentzLeft/2 + HWHM/2)

            lorentzian = yDoubleL

        elif self.is_singlet == False: #If it is a doublet


                for i in range(len(x)):

        

                        if round(x_values[i] - offset, 5) <= 0:
                            yDoubleL_right = 1 / (1 + np.power( (x_values[i] - offset)/ lorentzLeft, 2)) #/ (1 - self.branching_ratio) / np.pi / lorentzLeft/2

                        else:
                            
                            yDoubleL_right = 1 / (1 + np.power( (x_values[i] - offset)/ HWHM, 2)) #/ (1 - self.branching_ratio) / np.pi / HWHM/2


                        #Higher PE peak
                        if round(x_values[i] - offset - self.spinOrbitSplit, 5) <= 0:
                            yDoubleL_left = self.branching_ratio / (1 + np.power( (x_values[i] - (offset + self.spinOrbitSplit)) / lorentzLeft, 2)) #/ (1 - self.branching_ratio) / np.pi / lorentzLeft/2
                        else:
                            
                            yDoubleL_left = self.branching_ratio / (1 + np.power( (x_values[i] - (offset + self.spinOrbitSplit)) / HWHM, 2)) #/ (1 - self.branching_ratio) / np.pi / HWHM/2

                        yDoubleL[i] = (yDoubleL_right + yDoubleL_left) / (1 - self.branching_ratio) / np.pi



              

                lorentzian = yDoubleL

  
        doubleLorentz = scipy.signal.fftconvolve(gaussian,lorentzian,'same')
        doubleLorentz = doubleLorentz / (HWHM/4 + lorentzLeft/4 + HWHM_CK/4 + lorentzLeft_CK/4)
        scale = max(doubleLorentz)
        for i in range(len(doubleLorentz)):
            doubleLorentz[i] *= (self.amp/scale) #got rid of scale
        '''
        if PE_entry == True:
            x = x[::-1]
        '''
    
       


        self.peak_y = doubleLorentz
        
        return doubleLorentz





    #This is bad. Will fix later -Alaina
    def doniachSunjicFunc(self,x,PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        #Formula for Doniach-Sunjic peak equation:
        #
        #             cos(( pi  *  alpha ) / 2  +  (1 -  alpha ) * arctan((x - center) /  self.lorentz ))
        # Func =   ----------------------------------------------------------------------------
        #                                2          2  (1 - alpha) / 2
        #                  ( (x - center)   +   gamma )
        global photonEnergy
        self.photonEnergy = PE
        self.lorentz = width
        self.lorentz_CK = width_CK
        self.is_coster_kronig = coster_kronig
        self.gaussian = sigma
        self.amp = A
        self.asymmetry = asym
        self.asymmetryDoniach = asymD
        self.is_singlet = singlet
        self.branching_ratio = branch
        self.spinOrbitSplit = split

        data_range= max(x) - min(x)
        data_range /= 2

        middle = min(x)+data_range
        offset = self.photonEnergy-middle
        num_points = len(x)
        x_values, dx = np.linspace(-data_range,data_range,num_points,retstep=True)

        if self.gaussian ==0:
            self.gaussian += self.paramRange['Gaussian'][2]

        gaussian = np.exp(-np.power(x_values, 2) / ((np.power(self.gaussian, 2)))) / (self.gaussian * np.sqrt(np.pi))
        HWHM = self.lorentz/2
        HWHM_CK = self.lorentz/2
        if self.is_singlet == True:
            #self.asymmetryDoniach = -self.asymmetryDoniach #Opposite sign because it is in BE not KE


            #Singlet Doniach-Sunjic Equation:
            cos1 = (np.pi*self.asymmetryDoniach)/2
            cos2 = (1-self.asymmetryDoniach)*np.arctan((x_values+offset)/HWHM)
            numerator = np.cos(cos1 + cos2)
            de1 = pow(x_values+offset, 2) + pow(HWHM, 2)
            powDe = (self.asymmetryDoniach-1)/2
            denominator = pow(de1, powDe)
            lorentzian = numerator*denominator



        #Doesnt give doublet right now. Do not know why --> it follows the same methods as Aanalyzer
        elif self.is_singlet == False: #If it is a doublet

            if self.is_coster_kronig == True:

                cos1 = (np.pi*self.asymmetryDoniach)/2
                cos2 = (1-self.asymmetryDoniach)*np.arctan((x_values+offset)/HWHM)
                cos3 = (1-self.asymmetryDoniach)*np.arctan((x_values+self.spinOrbitSplit + offset)/HWHM_CK) #doublet only for the second part of the equation
                numerator1 = np.cos(cos1 + cos2)
                numerator2 = np.cos(cos1 + cos3)
                de1 = pow(x_values+offset, 2) + pow(HWHM, 2)
                de2 = pow(x_values + offset + self.spinOrbitSplit, 2) + pow(HWHM_CK, 2)
                powDe = (self.asymmetryDoniach-1)/2
                denominator1 = pow(de1, powDe)
                denominator2 = pow(de2, powDe)
                lorentzian = numerator1*denominator1 + self.branching_ratio*(numerator2*denominator2)

            else:

                cos1 = (np.pi*self.asymmetryDoniach)/2
                cos2 = (1-self.asymmetryDoniach)*np.arctan((x_values+offset)/HWHM)
                cos3 = (1-self.asymmetryDoniach)*np.arctan((x_values+self.spinOrbitSplit + offset)/HWHM) #doublet only for the second part of the equation
                numerator1 = np.cos(cos1 + cos2)
                numerator2 = np.cos(cos1 + cos3)
                de1 = pow(x_values+offset, 2) + pow(HWHM, 2)
                de2 = pow(x_values + offset + self.spinOrbitSplit, 2) + pow(HWHM, 2)
                powDe = (self.asymmetryDoniach-1)/2
                denominator1 = pow(de1, powDe)
                denominator2 = pow(de2, powDe)
                lorentzian = numerator1*denominator1 + self.branching_ratio*(numerator2*denominator2)




        doniachSunjic = scipy.signal.fftconvolve(gaussian,lorentzian,'same')

        #normalize the height so that intensity is the height of the max of the peak
        scale = max(doniachSunjic)
        for i in range(len(doniachSunjic)):
            doniachSunjic[i] *= (self.amp/scale)


        if x[0] < x[-1]: #KE CONDITION
            doniachSunjic = doniachSunjic[::-1]
        else:
            pass

        self.peak_y = doniachSunjic

        return doniachSunjic




#---------------------------Backgrounds Class---------------------------------------------------------------



class background(peak):
    """
    Background class is largely the same format as the peaks class,getY, set,get, function largely the same without
    most of the hassle the peaks come with

    One thing of note, most backgrounds are held by the individual, but SVSC backgrounds are held by a peak associated with them,
    and the peak takes on the management of it,but inside the background class it is the same as any other background

    To add a Background type:
    Add its bkgnType to the picker in the GUI
    Give its parameters ranges and pass them into the write_ini in the gui
    add the range to ini_parser and also the paramrange dict in xes_neo
    add that bkgnType in init here, and assign the function to self.bkgn
    add the get function for it
    add the set function in the gui xes_fit
    add the bgknType option to the self.bkgn_types at the top of xes_analysis2

    """
    def __init__(self,paramRange,bkgnType, peakType):
        self.bkgnType = bkgnType
        self.paramRange= paramRange
        self.peakType = peakType



        if self.bkgnType == 'Shirley-Sherwood':
            #self.bkgn = self.shirley_Sherwood



            #New Shirley Background
            self.bkgn = self.shirley_bkgn_again
            try:
                self.lorentzRange = np.arange(paramRange['Lorentzian'][0],paramRange['Lorentzian'][1],paramRange['Lorentzian'][2])
                self.photonEnergy = np.clip(self.photonEnergy,self.paramRange['Photon Energy'][0],self.paramRange['Photon Energy'][1])
                self.lorentz = np.random.choice(self.lorentzRange)
            except:
                self.lorentz = 1.2
                self.photonEnergy = 397.36
                
            
            if paramRange != '':
                self.k_range = np.arange(paramRange['k_range'][0],paramRange['k_range'][1],paramRange['k_range'][2])
                self.k = np.random.choice(self.k_range)
            else:
                self.k = 0
            
                
            '''
            if paramRange != '':
                self.backgroundShirRange = np.arange(paramRange['Shirley Background'][0],paramRange['Shirley Background'][1],paramRange['Shirley Background'][2])

                self.backgroundShirley = np.random.choice(self.backgroundShirRange)

            else:

                self.backgroundShirley = 100
            '''


        elif self.bkgnType.lower()  == 'linear':
            self.bkgn = self.linear_background
            try:
                self.backgroundRange = np.arange(paramRange['Background'][0],paramRange['Background'][1],paramRange['Background'][2])
                self.slopeRange = np.arange(paramRange['Slope'][0],paramRange['Slope'][1],paramRange['Slope'][2])

                #self.background is the b value in y = mx+b
                self.background = np.random.choice(self.backgroundRange)
                self.slope = np.random.choice(self.slopeRange)
                #self.slope = 0
            except:
                self.slope = 0
                self.background = -1
                if paramRange == '':
                    pass
        elif self.bkgnType == 'Exponential':
            self.bkgn = self.new_exponential

            try:
                self.A_range = np.arange(paramRange['Exponential Amplitude'][0],paramRange['Exponential Amplitude'][1],paramRange['Exponential Amplitude'][2])
                self.tau_range = np.arange(paramRange['Decay'][0],paramRange['Decay'][1],paramRange['Decay'][2])

                #self.background is the b value in y = mx+b
                self.A = np.random.choice(self.A_range)
                self.tau = np.random.choice(self.tau_range)
                #self.slope = 0
            except:
                self.A = 0
                self.tau = 0
                if paramRange == '':
                    pass
        elif self.bkgnType == 'Polynomial 1':
            self.bkgn = self.polynomial1
        elif self.bkgnType == 'Polynomial 2':
            self.bkgn = self.polynomial2
        elif self.bkgnType == 'Polynomial 3':
            self.bkgn = self.polynomial3
        elif self.bkgnType == '3-Param Tougaard':
            try:
                self.CTou3Range = np.arange(paramRange['CTou3'][0],paramRange['CTou3'][1],paramRange['CTou3'][2])
                self.DTou3Range = np.arange(paramRange['DTou3'][0],paramRange['DTou3'][1],paramRange['DTou3'][2])
                self.CTou3 = np.random.choice(self.CTou3Range)
                self.DTou3 = np.random.choice(self.DTou3Range)
            except:
                self.CTou3 = 1000 #Initial inputs in Aanalyzer
                self.DTou3 = 13300
            self.bkgn = self.Tougaard3Param
        elif self.bkgnType == '2-Param Tougaard':
            self.bkgn = self.Tougaard2Param
        elif self.bkgnType == 'SVSC_shirley':
            self.bkgn = self.shirley_Sherwood
            #func y vals tells you what to integrate over
            try:
                self.k_range = np.arange(paramRange['k_range'][0],paramRange['k_range'][1],paramRange['k_range'][2])
                self.k = np.random.choice(self.k_range)
            except:
                self.k = -1
                
        elif self.bkgnType == 'Baseline':
            self.bkgn = self.baseline2

            if paramRange != '':
                self.baselineRange = np.arange(paramRange['baseline'][0],paramRange['baseline'][1],paramRange['baseline'][2])
                self.baseline_value = np.random.choice(self.baselineRange)
            else:
                self.baseline_value = None

        else:
            print("Error Choosing Background in init of xes_fit")
            print("Background read as: " + str(self.bkgnType))
            exit()
        self.yBkgn = []

    def getY(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split,scale_var):
        y_val = []
        first = 1

        y_max = y.max()

        #Scaling is done proportional to the maximum peak height. This also sets background to 0
        if y_max <= 0.001:
            scale_val = first*100000000
        elif y_max <= 0.01:
            scale_val = first*10000000
        elif y_max <= 0.1:
            scale_val = first*1000000
        elif y_max <= 1:
            scale_val = first*100000
        elif y_max <= 10:
            scale_val = first*10000
        elif y_max <= 100:
            scale_val = first*1000
        elif y_max <= 1000:
            scale_val = first*100
        elif y_max <= 10000:
            scale_val = first*10
        else:
            scale_val = first  
            
        if scale_var == True:
            
            #y_val = y/first #Dividing every element by the first value 
            
            y_val = y*scale_val #Multiply by 1000 to scale it
               
        else:
            y_val = y
        y = y_val
       

        self.get_Background(x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)
        return self.yBkgn



    def get_Background(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        self.bkgn(x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)


    def mutate(self,chance):
        if self.bkgnType == 'Shirley-Sherwood':
            self.mutate_k(chance)
            self.mutate_backgroundShir
        if self.bkgnType == 'Baseline':
            self.mutate_baseline_value(chance)
        if self.bkgnType == 'Linear': #Added this in for new background --> Got rid of old slope
            self.mutate_background(chance)
            self.mutate_slope(chance)
        if self.bkgnType == 'SVSC_shirley':
            self.mutate_k(chance)
        if self.bkgnType == 'Exponential':
            self.mutate_A(chance)
            self.mutate_tau(chance)
        if self.bkgnType == '3-Param Tougaard':
            self.mutate_C(chance)
            self.mutate_D(chance)


    def mutate_k(self,chance):
        if random.random()*100 < chance:
            self.k = np.random.choice(self.k_range)

    #def mutate_backgroundShir(self,chance):
        #if random.random()*100 < chance:
            #self.backgroundShirley = np.random.choice(self.backgroundShirRange)

    def mutate_baseline_value(self,chance):
        if random.random()*100 < chance:
            self.baseline_value = np.random.choice(self.baselineRange)

    def mutate_background(self,chance):
        if random.random()*100 < chance:
            self.background = np.random.choice(self.backgroundRange)
    def mutate_slope(self,chance):
        if random.random()*100 < chance:
            self.slope = np.random.choice(self.slopeRange)

    def mutate_A(self,chance):
        if random.random()*100 < chance:
            self.A = np.random.choice(self.A_range)

    def mutate_tau(self,chance):
        if random.random()*100 < chance:
            self.tau = np.random.choice(self.tau_range)

    def mutate_C(self,chance):
        if random.random()*100 < chance:
            self.CTou3 = np.random.choice(self.CTou3Range)

    def mutate_D(self,chance):
        if random.random()*100 < chance:
            self.DTou3 = np.random.choice(self.DTou3Range)




    #Make sure to add in each background here
    def get(self):
       if self.bkgnType == 'Shirley-Sherwood':
            #return [self.k, self.backgroundShirley, self.bkgnType] #GET RID OF self.backgroundShirley
            return [self.k, self.bkgnType]
       elif self.bkgnType == 'Slope':
            return [self.bkgnType]
       elif self.bkgnType == 'Linear':
            return [self.new_b,self.slope,self.bkgnType]
       elif self.bkgnType == 'Exponential':
            return [self.A, self.tau, self.bkgnType]
       elif self.bkgnType == 'Baseline':
            return [self.baseline_value,self.bkgnType]
       elif self.bkgnType == 'Polynomial 1':
            return [self.bkgnType]
       elif self.bkgnType == 'Polynomial 2':
            return [self.bkgnType]
       elif self.bkgnType == 'Polynomial 3':
            return [self.bkgnType]
       elif self.bkgnType == '3-Param Tougaard':
            return [self.bkgnType, self.CTou3, self.DTou3]
       elif self.bkgnType == '2-Param Tougaard':
            return [self.bkgnType]
       elif self.bkgnType == 'SVSC_shirley':
           return [self.k,self.bkgnType]
    def getType(self):
        return self.bkgnType

    def set_k(self,newVal):
        self.k = newVal

    #def set_backgroundShir(self,newVal):
        #self.backgroundShirley = newVal

    def set_shirley_sherwood(self,params):
        self.k = params[0]
        #self.backgroundShirley = params[1]

    def set_baseline(self,params):
        self.baseline_value = params[0]


    def set_background(self,newVal):
        self.background = newVal

    def set_slope(self,newVal):
        self.slope = newVal

    def set_linear(self,params):
        self.new_b = params[0]
        self.slope = params[1]

    def set_exponential(self,params):
        self.A = params[0]
        self.tau = params[1]

    def set_Tougaard3(self,params):
        self.CTou3 = params[0]
        self.DTou3 = params[1]


    '''
    def baseline(self,x,y, PE, width, sigma, A, asym, asymD, singlet, branch, split):
        self.y = y
        self.x = x
        data_baseline = peakutils.baseline(x)
        funcs= data_baseline
        #bkgn_0 = data_baseline.modpoly(y, poly_order=0)[0]
        #funcs = bkgn_0
        self.yBkgn = funcs
        return self.yBkgn
    '''
    def baseline(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        self.y = y
        self.x = x

        poly_0 = np.polyfit(x, y, deg=0)
        funcs = np.polyval(poly_0, x)
        self.yBkgn = funcs

        return self.yBkgn



    def polynomial1(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split): #Doesnt work right now --> error in analysis x, y shapes (401), (0)
        self.y = y
        self.x = x

        numP = len(x)
        funcs = [0]*numP
        for i in np.arange(0, numP):
            funcs[i] = pow( (x[i]-x[int(np.rint(numP/2))]), 1);
        #poly_2 = np.polyfit(x, y, deg=2)
        #funcs = np.polyval(poly_2, x)
        self.yBkgn = funcs

        return self.yBkgn



    def polynomial2(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split): #Doesnt work right now --> error in analysis x, y shapes (401), (0)
        self.y = y
        self.x = x

        numP = len(x)
        funcs = [0]*numP
        for i in np.arange(0, numP):
            funcs[i] = pow( (x[i]-x[int(np.rint(numP/2))]), 2);
        #poly_2 = np.polyfit(x, y, deg=2)
        #funcs = np.polyval(poly_2, x)
        self.yBkgn = funcs

        return self.yBkgn



    def polynomial3(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        self.y = y
        self.x = x

        numP = len(x)
        funcs = [0]*numP
        for i in np.arange(0, numP):
            funcs[i] = pow( (x[i]-x[int(np.rint(numP/2))]), 3);

        self.yBkgn = funcs

        return self.yBkgn



    def new_exponential(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        numP = len(y)
        funcs =[0]*len(y)
        N = 10
        y_left = y[:N]
        y_0 = sum(y_left)/N #Average y at high PE
        x_0 = x[0]
        exp_0 = np.exp(-self.tau*x_0)
        pow_A = np.log10(exp_0)
        new_A = pow(self.A, pow_A)

        #print(self.A, self.tau)
        for i in np.arange(0, numP):
            #funcs[i] = y_0 -new_A*np.exp(-(x[i]*self.tau))
            #funcs[i] = y_0 -new_A*np.exp(-(x[i]*self.tau))
            #print(x[i], funcs[i])
            funcs[i] = y_0*np.exp(-self.tau*x[i])*np.exp(self.A)
            funcs[i] = y_0*np.exp(-self.tau*x[i])*np.exp(self.A)
        self.yBkgn = funcs

        return self.yBkgn


    def exponential_bkgn(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        self.y = y
        self.x = x

        #need to make these parameters outside background functions


        numP = len(y)
        funcs =[0]*len(y)

        #Taken from Aanalyzer code --> not sure why exponent is initially set to 1 or 0
        exponent = 1
        deltaExponent = max(abs(exponent / 100), 0.001)
        exponent += deltaExponent


        #Not sure if the x data needs to be flipped for PE instead of KE --> The exponential should be on the left side of the peak not the right
        for j in range(1, numP): #Cut off before numP so the end point is off. Need to fix this in order to scale down the righthand side of the background to the data
            gar = -exponent * (x[j] - x[numP // 2])

            if gar > 30:
                gar = 30
            elif gar < -30:
                gar = -30

            funcs[j] = -(np.exp(gar)) #Added negative sign to flip exponential to be in the -xy plane instead of +xy plane

        self.yBkgn = funcs

        return self.yBkgn



    def baseline2(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):

        bkgn_vals = [self.baseline_value]*len(y)
        self.yBkgn = bkgn_vals

        return bkgn_vals





     #Integral slope background works for now but is bad. Left side of data is not scaling properly
    def slope_bkgn(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        energyDelay= 0.5 #This is the initial value used in Aanalyzer --> not sure why??
        self.y = y
        self.x = x

        #need to make these parameters outside background functions
        numPeaksUsed = 1
        maxPeaks = 1
        numberBackgrounds = 1
        ma = maxPeaks + numberBackgrounds
        funcs = [0]*len(y)
        numP = len(y)



        for j in np.arange(numP-2, 0, -1): #error changed numP -1 to numP-2
            funcs[j] = (y[j] - y[numP-1]) * (x[j+1] - x[j]) + funcs[j+1] #changed y[numP] to y[numP -1] because of index error

        for j in range(1, numP):
            jDelay = 0
            x_eDelay = x[j] + energyDelay
            while j + jDelay < len(x) and x_eDelay > x[j + jDelay]:
                jDelay += 1
                if j + jDelay > numP:
                    jDelay -=1
                    break

            if j + jDelay > numP-1: #supposed to be just numP --> error
                funcs[j] = 0
            else:
                funcs[j] = funcs[j + jDelay]

        for j in np.arange(numP-2, 0, -1):#error changed numP -1 to numP-2
            funcs[j] = funcs[j] * (x[j+1] - x[j]) + funcs[j+1]


        self.yBkgn = funcs

        return self.yBkgn

    def new_slope(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):

        self.x = x
        self.y = y

        N = 10
        y_left = self.y[:N]
        y_right = self.y[N:]
        x_left = self.x[:N]
        x_right = self.x[N:]

        y_left_avg = sum(y_left)/N
        x_left_avg = sum(x_left)/N
        y_right_avg = sum(y_right)/N
        x_right_avg = sum(x_right)/N

        y_slope = (y_left_avg - y_right_avg)/(x_left_avg - x_right_avg)
        y_intercept = self.y[0] - y_slope*self.x[0]
        funcs = y_slope*self.x + y_intercept

        self.yBkgn = funcs

        return funcs



    def linear_background(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        self.y = y
        self.x = x
        numP = len(self.x)
        self.yBkgn = [0]*numP

        self.new_b = self.y[10] - self.slope*x[10]

        for i in range(numP):
            
            self.yBkgn[i] = self.linear(self.slope,self.x[i],self.new_b)
        #print(self.yBkgn)
        return self.yBkgn

    def linear(self,slope,x,new_b):
        
        return (slope*x)+new_b





    def better_shirley(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        #Not sure where this code came from. May need to cite it later. It has similar structure to the shirley formula described in Herrera's paper
        self.y = y
        self.x = x

        E = x
        J = y


        def integralOne(E, J, B, E1=0, E2=-1):
            integral = []
            if E2 < 0:
                E2 = len(J) + E2
            integral = sum([J[n] - B[n] for n in range(E,E2)])
            return integral

        def integralTwo(E, I, B, E1=0, E2=-1):
            integral = []
            if E2 < 0:
                E2 = len(I) + E2
            integral = sum([I[n] - B[n] for n in range(E1,E2)])
            return integral

        def getBn(E,I,B,E1=0,E2=-1):
            I2 = I[E2]
            I1 = I[E1]
            value = I2 + (I1 - I2)/(integralTwo(E,I,B,E1,E2))*integralOne(E,I,B,E1,E2)
            return value

        def iterateOnce(I,B,E1=0,E2=-1):
            b = [getBn(E,I,B,E1,E2) for E in range(len(I))]
            return b

        Bn = [0 for i in range(len(J))]
        Bn = iterateOnce(J,Bn)
        for i in range(6): #how many iterations it's doing
            B_temp = Bn
            Bn = iterateOnce(J,Bn)
            B_diff = [Bn[j] - B_temp[j] for j in range(len(Bn))] #Could make a check to see if the iterations are getting better. Usually little difference after 7 iterations

        self.yBkgn = Bn


        return self.yBkgn









    def SVSC_shirley(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):

        self.y = y
        self.x = x

        #Should probably declare these items outside of each background type
        numPeaksUsed = 1
        maxPeaks = 1
        numberBackgrounds = 1
        ma = maxPeaks + numberBackgrounds #will need to change to make it able to use multiple peaks/backgrounds
        numP = len(self.y)
        a =[0]*len(self.x) #dont know if we need a anymore?

        #voigt = peak.voigt
        funcs = y #len = numP -1 #recover initial peak curve --> How to get y points of just one peak??? Use PE range + some delta
        backgroundFromPeakShirleyFix = [0]*(len(self.y)-1) #not sure why its one less than the number of points
        SVSC_bkgn = backgroundFromPeakShirleyFix #easier to write --> original name comes from aanalyzer code

        a_old = 0.3
        a_new = 0.5 #are these initial values too large?
        old_fit = 10000
        best_fit = funcs #setting initial best fit --> just equal to y originally
        SVSC_diff = 1
        while a_new >= 0: #Iterates until a = 0, but keeps track of std of background to voigt fit. Need to find a better way for the GA to optimize a
            i = 1
            for i in range(maxPeaks):#calculates background for each peak then iterates
                #a_ratio is some parameter ratio --> I think it is the ratio of one parameter of different correlated peaks, unsure as to which parameter is being correlated
                a_ratio_b4 = a_old #Right now these are just random --> real code: a[ peakShirleyma[ peakShirelyCorrTo] ] / a[ mama[peakShirelyCorrTo] ]
                a_ratio_after = a_new #defined on line 15233 in PUnit1 --> Values are a[] before and after lfitmod is called
                peakShirleyBackground = 0.8*a_ratio_b4 + 0.2*a_ratio_after #I think this is supposed to be the scattering factor? Now sure how it is optimized
                #Maybe for now we should treat peakShirleyBackground as the scattering factor?

                for j in np.arange(numP -2, 0, -1):
                    SVSC_bkgn[j-1] = self.y[j-1]*-(self.x[j+1]-self.x[j])*peakShirleyBackground + SVSC_bkgn[j] #isnt this just what we already had but now with a wider range?
                    funcs[j] += SVSC_bkgn[j-1]
                #should write array in here to store each peak curves background --> will sum these up later
                i +=1

                iteration_diff = np.subtract(voigt, funcs) #need to change voigt to whatever the curve fit y array is
                new_fit = np.std(iteration_diff)
                new_fit_array = funcs
                if new_fit < old_fit:
                    old_fit = new_fit
                    best_fit = new_fit_array
                a_old = a_new
                a_new -= 0.01 #slow decrease for now --> NEED TO FIND BETTER WAY TO OPTIMIZE a_new
                #lfitmod caluculated here --> Calcualtes parameters between iterations: This is what makes the background active
                #Should we call class Peak here to recalculate the fit with the new background? Active curve fitting

        funcs = best_fit
        #return funcs #Not sure how we are calling this (self.yBKgn?)
        self.yBkgn = funcs

        return self.yBkgn





    def shirley_Sherwood(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        #too lazy to find all the x and ys and get rid of the self
        self.y = y
        self.x = x

        #need to make these parameters outside background functions
        numPeaksUsed = 1
        maxPeaks = 1
        numberBackgrounds = 1
        ma = maxPeaks + numberBackgrounds
        useIntegralBkgn=True
        numP = [0]*len(self.y) #we are using this as an array right now but it should just be the number of data points
        a =[0]*len(self.x)


        def iterations(self,x,y):
            numPeaksUsed = 1
            maxPeaks = 1
            numberBackgrounds = 1
            ma = maxPeaks + numberBackgrounds
            useIntegralBkgn=True
            numP = [0]*len(self.y) #we are using this as an array right now but it should just be the number of data points
            a =[0]*len(self.x)
            #need this to find the correct data points in which the bakcground will be removed
            numPointsAroundBackgroundLimitsLocal = 5
            nRightLocal = numPointsAroundBackgroundLimitsLocal // 2
            nLeftLocal = numPointsAroundBackgroundLimitsLocal // 2

            yRightLocal = 0
            yLeftLocal = 0

            for j in range(-(numPointsAroundBackgroundLimitsLocal // 2), numPointsAroundBackgroundLimitsLocal // 2 + 1):
                #yRightLocal += datos[dataNumber].ModifiedCurve.y[nRightLocal + j]
                #yLeftLocal += datos[dataNumber].ModifiedCurve.y[nLeftLocal + j]
                yRightLocal += self.y[len(self.y) - nRightLocal-1 + j]
                yLeftLocal += self.y[nLeftLocal + j]

            yLeftLocal /= (numPointsAroundBackgroundLimitsLocal // 2) * 2 + 1
            yRightLocal /= (numPointsAroundBackgroundLimitsLocal // 2) * 2 + 1



            nLeft = nLeftLocal
            nRight = len(x)-nRightLocal
            global yRight
            yRight = yRightLocal
            yLeft = yLeftLocal


            iterationsIntegralBkgn = 6
            BkgdCurve = []
            funcs = numP
            #funcs = np.zeros((ma, numP))

            if useIntegralBkgn: #from Aanalyzer code line 14867
                ma += 1 #Need to add in array to store each peak background peak[ma] --> sum in other backgrounds? peak[ma] += funcs...
                #funcs[ma][numP] = 0
                #print("K is " + str(self.k))
                for j in range(nRight-1, -1, -1):
                    #funcs[ma][j] = (self.y[j] - yRight[j]) * (self.x[j+1] - self.x[j]) + funcs[ma][j+1]
                    #print(self.y[j] - yRight)
                    funcs[j] = (self.y[j] - yRight) *self.k* -(self.x[j+1] - self.x[j]) + funcs[j+1] #assumes x is in KE, not sure if that changes anything

                '''
                for j in range(0, nLeft):
                    #funcs[ma][j] = funcs[ma][nLeft]
                    funcs[j] = yLeft-yRight
                '''
                integralma = ma

            '''
            #iterates shirley background
            if useIntegralBkgn: #from Aanalyzer code line 15140
                for l in range(iterationsIntegralBkgn):
                    for j in range(nRight-1, nLeft, -1):
                        #funcs[integralma][j] = (self.y[j] - yRight[j] - a[integralma] * funcs[integralma][j]) * (self.x[j+1] - self.x[j]) + funcs[integralma][j+1]
                        funcs[j] = (self.y[j] - yRight - funcs[j]) * (self.x[j+1] - self.x[j]) + funcs[j+1]
                    for j in range(1, nLeft):
                        #funcs[integralma][j] = funcs[integralma][nLeft]
                        funcs[j] = funcs[nLeft]
                        #calls lfitmod here -->calculates chisq and deletes all parameters

                    l += 1
            '''
            return funcs
        for i in range(1): #How many iterations it is performing
            funcs = iterations(self,x,y)
        self.yBkgn = funcs

        ''' Old built in baseline (bad)
        for i in range(len(self.yBkgn)):
            self.yBkgn[i] += yRight

        return self.yBkgn
        '''
    '''
    Just barely started on peak shirley, commented out so it wont cause a compilation error
    def peak_shirley(self,x,y,peak):
        peak.getY
    '''

    @nb.jit
    def Tougaard3Param(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        funcs = np.zeros(len(y))
        CTou3 = self.CTou3 #1000 #These values come from Aanalyzer --> not sure why these are the initial ones but it allows for the user to change them
        DTou3 = self.DTou3
        numP = len(y) #we are using this as an array right now but it should just be the number of data points

        #need this to find the correct data points in which the bakcground will be removed
        numPointsAroundBackgroundLimitsLocal = 5
        nRightLocal = numPointsAroundBackgroundLimitsLocal // 2
        nLeftLocal = numPointsAroundBackgroundLimitsLocal // 2

        yRightLocal = 0
        yLeftLocal = 0

        for j in range(-(numPointsAroundBackgroundLimitsLocal // 2), numPointsAroundBackgroundLimitsLocal // 2 + 1):
            yRightLocal += y[len(y) - nRightLocal-1 + j]
            yLeftLocal += y[nLeftLocal + j]

        yLeftLocal /= (numPointsAroundBackgroundLimitsLocal // 2) * 2 + 1
        yRightLocal /= (numPointsAroundBackgroundLimitsLocal // 2) * 2 + 1

        nRightLocal = len(x)-nRightLocal


        for i in np.arange(nLeftLocal, nRightLocal):
            integralGar = 0
            for j in np.arange(i, nRightLocal):
                energyDif = -(x[j] - x[i])
                integralGar += (y[j] - yRightLocal)*energyDif/(pow(CTou3-pow(energyDif,2),2) + ((DTou3*pow(energyDif,2))*-(x[j+1]-x[j])))
            funcs[i] = integralGar

        BTou3 = (yLeftLocal - yRightLocal)/funcs[nLeftLocal]
        for i in np.arange(nLeftLocal, nRightLocal):
            funcs[i] *= BTou3
            funcs[i] += yRightLocal

        for i in np.arange(0, nLeftLocal):
            funcs[i] = yLeftLocal

        for i in np.arange(nRightLocal, numP):
            funcs[i] = yRightLocal
            self.yBkgn = funcs

        return self.yBkgn





    @nb.jit
    def Tougaard2Param(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split): #Added in but does not work right now
        funcs = np.zeros(len(y))
        numP = len(y) #we are using this as an array right now but it should just be the number of data points

        #need this to find the correct data points in which the bakcground will be removed
        numPointsAroundBackgroundLimitsLocal = 5
        nRightLocal = numPointsAroundBackgroundLimitsLocal // 2
        nLeftLocal = numPointsAroundBackgroundLimitsLocal // 2

        yRightLocal = 0
        yLeftLocal = 0

        for j in range(-(numPointsAroundBackgroundLimitsLocal // 2), numPointsAroundBackgroundLimitsLocal // 2 + 1):
            yRightLocal += y[len(y) - nRightLocal-1 + j]
            yLeftLocal += y[nLeftLocal + j]

        yLeftLocal /= (numPointsAroundBackgroundLimitsLocal // 2) * 2 + 1
        yRightLocal /= (numPointsAroundBackgroundLimitsLocal // 2) * 2 + 1

        nRightLocal = len(x)-nRightLocal

        CTou2 = 1643 #Taken from Aanalyzer



        for i in np.arange(nLeftLocal, nRightLocal):
            integralGar = 0
            for j in np.arange(i, nRightLocal):
                energyDif = -(x[j] - x[i])
                integralGar += (y[j] - yRightLocal)*energyDif/CTou2-pow(energyDif,2)/CTou2-pow(energyDif,2)*(x[j+1] - x[j])
            funcs[i] = integralGar

        BTou2 = (yLeftLocal - yRightLocal)/funcs[nLeftLocal]
        for i in np.arange(nLeftLocal, nRightLocal):
            funcs[i] *= BTou2
            funcs[i] += yRightLocal

        for i in np.arange(0, nLeftLocal):
            funcs[i] = yLeftLocal

        for i in np.arange(nRightLocal, numP):
            funcs[i] = yRightLocal
            self.yBkgn = funcs

        return self.yBkgn




    def shirley_bkgn_again(self,x,y, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split):
        

        #Dont want to be dependent upon I1 and I2:
        #Change to be more like:
        #
        #
        #S(E) = baseline_shirley + k1*(intensity integral peak 1)*Area peak 1 + k2*(intensity integral peak 2)*Area peak 2 .... for each peak
        #
        #Each peak is given a unique scattering factor valued between 0 and 1
        self.y = y
        self.x = x
        #integral = [0]*len(self.y) #we are using this as an array right now but it should just be the number of data points\
        #integral = [self.backgroundShirley]*len(y)


        peak_class = peak(self.paramRange, self.peakType)
        #baseline_shirley = [y[-1]]*len(y) #First value is just last data point in y array aka the rightside of data
        #self.photonEnergy = peak_class.photonEnergy

        if(self.peakType == "Voigt"):


            self.yValues = peak_class.voigtFunc(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split) #These are not the same as the ones we get from peak.peakFunc --> Object attribute error

        elif(self.peakType == "Gaussian"):
            self.yValues = peak_class.gaussFunc(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)
        elif(self.peakType == "Lorentzian"):
            self.yValues = peak_class.lorentzFunc(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)
        elif(self.peakType == "Double Lorentzian"):
            self.yValues = peak_class.doubleLorentzFunc(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)

        elif(self.peakType == "Doniach-Sunjic"):
            self.yValues = peak_class.doniachSunjicFunc(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)

        else:
            print("Error assigning peak type")
            print("Peaktype found is: " + str(self.peakType))
            exit()

        baseline_shirley = [min(self.yValues)]*len(y) #Using the smallest y value as the basis for the right side of the data
        yValues = self.yValues

        '''
        for i in range(len(x)):
            print(x[i], self.yValues[i])
        '''

        def areaA(baseline_shirley):

            numPointsAroundBackgroundLimitsLocal = 5
            nRightLocal = numPointsAroundBackgroundLimitsLocal // 2


            yRightLocal = 0


            for j in range(-(numPointsAroundBackgroundLimitsLocal // 2), numPointsAroundBackgroundLimitsLocal // 2 + 1):
                yRightLocal += self.y[len(self.y) - nRightLocal-1 + j]
            #Upper bound of integral
            yRightLocal /= (numPointsAroundBackgroundLimitsLocal // 2) * 2 + 1
            nRight = len(x)-nRightLocal
            A_sum = 0
            areaA = np.zeros(len(y))
            area_A = np.zeros(len(y))
            #print(nRight)
            #print(nRight)
            for i in np.arange(nRight-1, -1, -1):
                A_sum = np.sum(areaA)

                #areaA[i] = (((self.yValues[i]-baseline_shirley[i])+(self.yValues[i+1]-baseline_shirley[i+1]))/2)*(x[i+1]-x[i])
                areaA[i] = (((self.yValues[i])+(self.yValues[i+1])/2)*(x[i+1]-x[i]))
                area_A[i] = A_sum
                '''
                if areaA[i +1] > areaA[i]:
                    area_A[i] = A_sum - areaA[i]
                elif areaA[i + 1] == areaA[i]:
                    area_A[i] = area_A[i + 1]
                else:
                    pass
                '''

            return abs(area_A)

        total_area = areaA(baseline_shirley)






        def iterations(self,x,baseline_shirley, k_val, yValues):

            #need this to find the correct data points in which the bakcground will be removed
            baseline_shirley = [min(yValues)]*len(y)
            numPointsAroundBackgroundLimitsLocal = 5
            nRightLocal = numPointsAroundBackgroundLimitsLocal // 2


            yRightLocal = 0


            for j in range(-(numPointsAroundBackgroundLimitsLocal // 2), numPointsAroundBackgroundLimitsLocal // 2 + 1):
                yRightLocal += self.y[len(self.y) - nRightLocal-1 + j]
            #Upper bound of integral
            yRightLocal /= (numPointsAroundBackgroundLimitsLocal // 2) * 2 + 1
            nRight = len(x)-nRightLocal
            yRight = yRightLocal

            

            integral = baseline_shirley

            new_integral = baseline_shirley
            deltaX = abs(x[1] - x[2]) #Data stepsize --> this is assumed that the x data is stepwise and equal throughout the whole data range
            #Integral Calculation

            for j in range(nRight-1, -1, -1): #How to include area into this equation??

                #integral[j] = integral[j + 1] + k_val*(yValues[j] - baseline_shirley[j])*-(self.x[j+1] - self.x[j])#yValues and baseline_shirley should change with each
                #yRight is the avg of the right side of the data around the peal
                #k is the scattering factor valued bewteen 0 and 1
                #yValues - baseline_shirley is the difference from the curve fit to the background with baseline_shirley taking on the previous integral fit values for each iteration
                new_integral[j] = integral[j] + k_val*total_area[j]
                #print(k_val*total_area[j])
                #print(k_val*total_area[j])
            #print(np.array(new_integral).sum(), np.array(integral).sum())
            #return integral
            return new_integral




        self.yBkgn = iterations(self,x, baseline_shirley, self.k, yValues)


        return self.yBkgn



from tkinter import *
import time
from threading import *

# Create Object
root = Tk()
var = "bobthebuilder"
# Set geometry
root.geometry("400x400")

# use threading
def doSomething():
    t1=Thread(target=printanything)
    print("Hello")
    t1.start()

def printanything():
    global var
    var = "new"

def threading():
    # Call work function
    t1=Thread(target=work)
    t1.start()

# work function
def work():

    print("sleep time start")

    for i in range(10):
        global var
        print(var)
        time.sleep(1)

    print("sleep time stop")

# Create Button
Button(root,text="Click Me",command = threading).pack()
Button(root, text="Anythin", command=doSomething).pack()

# Execute Tkinter
root.mainloop()
